package modules.render;

import java.util.ArrayList;

import org.lwjgl.opengl.Display;
import org.lwjgl.opengl.GL11;
import org.lwjgl.opengl.GL13;
import org.lwjgl.opengl.GL30;

import utility.BufferConverter;
import utility.Timer;

import modules.Module;
import modules.commands.CommandExecutor;
import modules.commands.Command_Factory;
import modules.loader.Scene;
import modules.loader.loader.ViewportXMLSave;
import modules.render.cpu.CPUIllum;
import modules.render.objects.ICamera;
import modules.render.objects.IDirectionalLight;
import modules.render.objects.IRenderable;


public class RenderModule extends Module {

	private static boolean DRAWDEBUG = false;
	private static boolean DRAWCOMPOSITE = true;
	
	private boolean rotateCam = false;
	
	private boolean useCPU = false;
	private int counter = 0;
	private boolean finishedCPU = false;
	
	private boolean drawnToFile = true; // if set true - never written
	private boolean drawScene = true;	// if set true - generates scene from level file
	
	private boolean loadedFromFile;
	
	private WindowWrapper window;
	private Timer timer;
	
	private TextureCollection textures;
	private FBOCollection fbos;
	private VBO vbo;
	private Shader defaultTexturedShader;
	private Shader splitAtlasShader;
	private Shader gBufferShader;
	private Shader lighting;
	private Shader lightingLights;
	private Shader shadow;
	
	private int cpuResultIndex;
	private CPUIllum cpuIllum;
	
	int albedo;
	int normal;
	int depth;
	int light;
	
	public RenderModule(Scene scene, CommandExecutor commandExecutor, Command_Factory commandFactory, int updateFrequencyMs) {
		super(scene, commandExecutor, commandFactory, updateFrequencyMs);
		
		window = new WindowWrapper(commandExecutor, commandFactory.shutDownCommand());
		timer = new Timer();
		
		cpuIllum = new CPUIllum();
	}

	@Override
	protected void init() {
		window.initializeDisplay();
		
		GL11.glEnable(GL11.GL_TEXTURE_2D);
		GL11.glEnable(GL11.GL_DEPTH_TEST);
		
		textures = new TextureCollection();
		fbos = new FBOCollection(WindowWrapper.WINDOWX, WindowWrapper.WINDOWY);
		vbo = new VBO();
		
		// ----- SHADER INIT -----
		defaultTexturedShader = new Shader("/content/shader/defaultTextured");
		defaultTexturedShader.initUniform("colorMap");
		
		gBufferShader = new Shader("/content/shader/gBuffer");
		gBufferShader.initUniform("viewMatrix");
		gBufferShader.initUniform("projMatrix");
		gBufferShader.initUniform("lightColor");
		
		splitAtlasShader = new Shader("/content/shader/splitAtlas");
		splitAtlasShader.initUniform("index");
		splitAtlasShader.initUniform("gridDimension");
		splitAtlasShader.initUniform("atlas");
		
		shadow = new Shader("/content/shader/shadow");
		shadow.initUniform("viewMatrix");
		shadow.initUniform("projMatrix");
		for (int i = 0; i < Scene.NUMLIGHTS; i++) {
			shadow.initUniform("lightView["+i+"]");
		}
		shadow.initUniform("lightProj");	
		shadow.initUniform("shadowAtlas");
		
		lighting = new Shader("/content/shader/lighting");
		lighting.initUniform("viewMatrix");
		lighting.initUniform("projMatrix");
		lighting.initUniform("invViewMatrix");
		lighting.initUniform("invProjMatrix");
		for (int i = 0; i < Scene.NUMLIGHTS+1; i++) {
			lighting.initUniform("lightView["+i+"]");
		}
		lighting.initUniform("lightProj");	
		for (int i = 0; i < Scene.NUMLIGHTS+1; i++) {
			lighting.initUniform("albedoAtlas["+i+"]");
		}
		for (int i = 0; i < Scene.NUMLIGHTS+1; i++) {
			lighting.initUniform("normalAtlas["+i+"]");
		}
		for (int i = 0; i < Scene.NUMLIGHTS+1; i++) {
			lighting.initUniform("depthAtlas["+i+"]");
		}
		for (int i = 0; i < Scene.NUMLIGHTS+1; i++) {
			lighting.initUniform("lightAtlas["+i+"]");
		}
		lighting.initUniform("gridDimension");
		
		lightingLights = new Shader("/content/shader/lightingLights");
		lightingLights.initUniform("viewMatrix");
		lightingLights.initUniform("projMatrix");
		lightingLights.initUniform("invViewMatrix");
		lightingLights.initUniform("invProjMatrix");
		for (int i = 0; i < Scene.NUMLIGHTS; i++) {
			lightingLights.initUniform("lightView["+i+"]");
		}
		lightingLights.initUniform("lightProj");	
		lightingLights.initUniform("albedoAtlas");
		lightingLights.initUniform("normalAtlas");
		lightingLights.initUniform("depthAtlas");
		lightingLights.initUniform("lightAtlas");
		for (int i = 0; i < Scene.NUMLIGHTS; i++) {
			lightingLights.initUniform("lightAtlas["+i+"]");
		}
		lightingLights.initUniform("gridDimension");
		// ----- SHADER END -----
		
		if(!drawScene) loadedFromFile = false; // if set true - never load
		
	}
	
	/**
	 * Returns true if at least some objects were drawn
	 * 
	 * @param lights
	 * @return
	 */
	private boolean defaultDraw(ArrayList<IDirectionalLight> lights) {

		// upload scene objects if necessary
		ArrayList<IRenderable> objects = scene.getRenderObjects();
		for (int i = 0; i < objects.size(); i++) {
			if (!objects.get(i).vertexDataIsUploaded()) {
				objects.get(i).uploadVertexData(vbo);
				objects.get(i).uploadTexture(textures);
			}
		}
		
		//TODO DEBUG SPHERES
		/*ArrayList<IRenderable> debugSpheres = scene.getDebugSpheres();
		for (int i = 0; i < debugSpheres.size(); i++) {
			if (!debugSpheres.get(i).vertexDataIsUploaded()) {
				debugSpheres.get(i).uploadVertexData(vbo);
				debugSpheres.get(i).uploadTexture(textures);
			}
		}*/
		
		fbos.bindAtlasRegion(0, 0, 0);
		GL11.glClear(GL11.GL_COLOR_BUFFER_BIT | GL11.GL_DEPTH_BUFFER_BIT);
		
		// clear light background
		// GL11.glClearColor(0.95f, 0.95f, 0.95f, 1.0f);
		GL30.glClearBuffer(GL11.GL_COLOR, 2, BufferConverter.toBuffer(new float[]{0.0f, 0.0f, 0.0f, 1.0f})); // TODO (commented out generates a correct result now...) - ATTENTION - location is directly 0, 1, 2 (not a built in variable!)
		
		GL11.glEnable(GL11.GL_CULL_FACE); 
		//GL11.glCullFace(GL11.GL_FRONT); //TODO - temporary disabled backface drawing for shadows
		
		// draw all lights into different atlas regions
		for (int l = 0; l < lights.size()+1; l++) {
			
			fbos.bindAtlasRegion(0, 0, l);
			
			// WHAT SHADER TO BIND? (reduce shader binds?)
			gBufferShader.bind();
			
			if (l == 0) {
				gBufferShader.setValueMat4f("viewMatrix", BufferConverter.toBuffer(scene.getCamera().getViewMatrix()));
				gBufferShader.setValueMat4f("projMatrix", BufferConverter.toBuffer(scene.getCamera().getProjMatrix()));
			} else {
				//camera
				gBufferShader.setValueMat4f("viewMatrix", BufferConverter.toBuffer(lights.get(l-1).getViewMatrix()));
				gBufferShader.setValueMat4f("projMatrix", BufferConverter.toBuffer(lights.get(l-1).getProjMatrix()));
				/*if ((l%3) == 0) {
					gBufferShader.setValueVec3f("lightColor", 1.0f, 0.0f, 0.0f);
				} else if ((l%3) == 1) {
					gBufferShader.setValueVec3f("lightColor", 0.0f, 1.0f, 0.0f);
				} else if ((l%3) == 2) {
					gBufferShader.setValueVec3f("lightColor", 0.0f, 0.0f, 1.0f);
				} else {*/
					gBufferShader.setValueVec3f("lightColor", 1.0f, 1.0f, 1.0f);
				//}
			}
			
			// draw
			for (int i = 0; i < objects.size(); i++) {
				if (objects.get(i).vertexDataIsUploaded()) {
					objects.get(i).draw(gBufferShader, vbo, textures);
				}
			}
			/*//TODO DEBUG SPHERES
			GL11.glPolygonMode( GL11.GL_FRONT_AND_BACK, GL11.GL_LINE );
			for (int i = 0; i < debugSpheres.size(); i++) {
				if (debugSpheres.get(i).vertexDataIsUploaded()) {
					debugSpheres.get(i).draw(gBufferShader, vbo, textures);
				}
			}
			GL11.glPolygonMode( GL11.GL_FRONT_AND_BACK, GL11.GL_FILL );*/
			
			gBufferShader.unbind();
			
			fbos.unbind();
			
		}
		
		//GL11.glCullFace(GL11.GL_BACK); //TODO - temporary disabled backface drawing for shadows
		
		if (objects.size() == 0) {
			return false;
		} else {
			return true;
		}
		
	}
	
	@Override
	protected void update() {
		
		timer.waitFor(100);
		
		timer.startMeasure();
		
		window.updateResize();
		window.updateKeys();
		
		if (rotateCam) scene.getCamera().rotateY(25.0f*(float)timer.dt());
		
		GL11.glClearColor(0.95f, 0.95f, 0.95f, 1.0f);
		GL11.glClear(GL11.GL_COLOR_BUFFER_BIT | GL11.GL_DEPTH_BUFFER_BIT);
		GL11.glViewport(0, 0, WindowWrapper.WINDOWX, WindowWrapper.WINDOWY);
		
		// upload SCREEN-QUAD if necessary
		ArrayList<IRenderable> screenQuad = scene.getScreenQuadObjects();
		for (int i = 0; i < screenQuad.size(); i++) {
			if (!screenQuad.get(i).vertexDataIsUploaded()) {
				screenQuad.get(i).uploadVertexData(vbo);
				screenQuad.get(i).uploadTexture(textures);
			}
		}
		
		//GL11.glDisable(GL11.GL_CULL_FACE);
		//GL11.glColor3f(1.0f, 1.0f, 0.0f);
		
		ArrayList<IDirectionalLight> lights = scene.getLights();
		
		
		boolean objectsWereDrawn = false;
		
		if (drawScene) {
			// ---------------- DRAW LIGHT PERSPECTIVES ----------------
			objectsWereDrawn = defaultDraw(lights);
			// ---------------- DRAW LIGHTS END ----------------
		} else {
		
			// load from file if necessary
			if (!loadedFromFile) {
				
				ArrayList<ICamera> viewports = new ArrayList<ICamera>();
						
				try {
					viewports = ViewportXMLSave.loadObjectDataFromXML();
				} catch (Exception e) {
					e.printStackTrace();
				}
				
				// update matrices
				scene.getCamera().setViewMatrix(viewports.get(0).getViewMatrix());
				scene.getCamera().setProjMatrix(viewports.get(0).getProjMatrix());
				scene.getCamera().setInvProjMatrix(viewports.get(0).getInvProjMatrix());
				
				for (int i = 0; i < viewports.size()-1; i++) {
					scene.getLights().get(i).setViewMatrix(viewports.get(1+i).getViewMatrix());
					scene.getLights().get(i).setProjMatrix(viewports.get(1+i).getProjMatrix());
					scene.getLights().get(i).setInvProjMatrix(viewports.get(1+i).getInvProjMatrix());
				}
				
				// update images
				albedo = textures.uploadImage("/content/output/albedo_inv.png");
				normal = textures.uploadImage("/content/output/normal_inv.png");
				depth = textures.uploadImage("/content/output/depth_inv.png");
				light = textures.uploadImage("/content/output/light_inv.png");
				
				objectsWereDrawn = true;
				loadedFromFile = true;
			}
			
		}

		GL11.glViewport(0, 0, WindowWrapper.WINDOWX, WindowWrapper.WINDOWY);
		
		// ---------------- DRAW TO SUB FBO ----------------
		
		for (int j=0; j<(lights.size()+1); j++) { //lights.size()
			
			// albedo
			fbos.bindAll(1 + j, 0);
			fbos.clear(1 + j, 0);
			if (drawScene)
				fbos.bind_texture(0, 0, 0);
			else
				textures.bindTexture(albedo);
			splitAtlasShader.bind();
			splitAtlasShader.setValue1i("index", j);
			splitAtlasShader.setValueVec2f("gridDimension", Scene.GRIDDIMENSION[0], Scene.GRIDDIMENSION[1]);
			splitAtlasShader.setValue1i("atlas", 0);
			// draw (single quad normally)
			for (int i = 0; i < screenQuad.size(); i++) {
				if (i == 0 && screenQuad.get(i).vertexDataIsUploaded()) {
					screenQuad.get(i).draw(splitAtlasShader, vbo, textures);
				}
			}
			splitAtlasShader.unbind();
			fbos.unbind_texture();
			fbos.unbind();
			
			fbos.genMipMaps(1 + j, 0);
			
			// normal (light, depth)
			fbos.bindSingle(1 + j, 1, 0); //1 + lights.size()+1 + j, 0
			fbos.clear(1 + j, 1);
			if (drawScene)
				fbos.bind_texture(0, 0, 1);
			else
				textures.bindTexture(normal);
			splitAtlasShader.bind();
			splitAtlasShader.setValue1i("index", j);
			splitAtlasShader.setValueVec2f("gridDimension", Scene.GRIDDIMENSION[0], Scene.GRIDDIMENSION[1]);
			splitAtlasShader.setValue1i("atlas", 0);
			// draw (single quad normally)
			for (int i = 0; i < screenQuad.size(); i++) {
				if (i == 0 && screenQuad.get(i).vertexDataIsUploaded()) {
					screenQuad.get(i).draw(splitAtlasShader, vbo, textures);
				}
			}
			splitAtlasShader.unbind();
			fbos.unbind_texture();
			fbos.unbind();
			
			fbos.genMipMaps(1 + j, 1);
			
			// light (depth)
			fbos.bindSingle(1 + j, 2, 0);
			fbos.clear(1 + j, 2);
			if (drawScene)
				fbos.bind_texture(0, 0, 2);
			else
				textures.bindTexture(light);
			splitAtlasShader.bind();
			splitAtlasShader.setValue1i("index", j);
			splitAtlasShader.setValueVec2f("gridDimension", Scene.GRIDDIMENSION[0], Scene.GRIDDIMENSION[1]);
			splitAtlasShader.setValue1i("atlas", 0);
			// draw (single quad normally)
			for (int i = 0; i < screenQuad.size(); i++) {
				if (i == 0 && screenQuad.get(i).vertexDataIsUploaded()) {
					screenQuad.get(i).draw(splitAtlasShader, vbo, textures);
				}
			}
			splitAtlasShader.unbind();
			fbos.unbind_texture();
			fbos.unbind();
			
			fbos.genMipMaps(1 + j, 2);
			
			// depth
			//GL11.glViewport(0, 0, 2*WindowWrapper.WINDOWX, 2*WindowWrapper.WINDOWY);
			fbos.bindSingle(1 + j, 3, 0);
			fbos.clear(1 + j, 3);
			if (drawScene)
				fbos.bind_texture(0, 0, 3);
			else
				textures.bindTexture(depth);
			splitAtlasShader.bind();
			splitAtlasShader.setValue1i("index", j);
			splitAtlasShader.setValueVec2f("gridDimension", Scene.GRIDDIMENSION[0], Scene.GRIDDIMENSION[1]);
			splitAtlasShader.setValue1i("atlas", 0);
			// draw (single quad normally)
			for (int i = 0; i < screenQuad.size(); i++) {
				if (i == 0 && screenQuad.get(i).vertexDataIsUploaded()) {
					screenQuad.get(i).draw(splitAtlasShader, vbo, textures);
				}
			}
			splitAtlasShader.unbind();
			fbos.unbind_texture();
			fbos.unbind();
			//GL11.glViewport(0, 0, WindowWrapper.WINDOWX, WindowWrapper.WINDOWY);
			
			fbos.genMipMaps(1 + j, 3);
			
			//GL11.glViewport(0, 0, 2*WindowWrapper.WINDOWX, 2*WindowWrapper.WINDOWY);
			// depth 2
			fbos.bindSingle(1 + j, 5, 0);
			fbos.clear(1 + j, 5);
			if (drawScene)
				fbos.bind_texture(0, 0, 3);
			else
				textures.bindTexture(depth);
			splitAtlasShader.bind();
			splitAtlasShader.setValue1i("index", j);
			splitAtlasShader.setValueVec2f("gridDimension", Scene.GRIDDIMENSION[0], Scene.GRIDDIMENSION[1]);
			splitAtlasShader.setValue1i("atlas", 0);
			// draw (single quad normally)
			for (int i = 0; i < screenQuad.size(); i++) {
				if (i == 0 && screenQuad.get(i).vertexDataIsUploaded()) {
					screenQuad.get(i).draw(splitAtlasShader, vbo, textures);
				}
			}
			splitAtlasShader.unbind();
			fbos.unbind_texture();
			fbos.unbind();
			
			fbos.genMipMaps(1 + j, 5);
			//GL11.glViewport(0, 0, WindowWrapper.WINDOWX, WindowWrapper.WINDOWY);
		}
		
		if (DRAWDEBUG) {
		
			drawDebug(screenQuad);
			
		}
		
		if (!useCPU && DRAWCOMPOSITE) {
			// ---------------- DRAW SCENE ----------------
			
			fbos.bindAll(1, 4); // camera scene // was 1 + 4*(lights.size()+1) + 0
			fbos.clear(1, 4);
			
			lighting.bind();
			
			/*GL13.glActiveTexture(GL13.GL_TEXTURE0);
			fbos.bind_texture(0, 0, 1);
			lighting.setValue1i("normalAtlas", 0);*/
			
			int lightsNum = lights.size()+1;
			// albedo maps // Interesting result if using normals instead of light...
			for (int i = 0; i < lightsNum; i++) {
				GL13.glActiveTexture(GL13.GL_TEXTURE0+i);
				fbos.bind_texture(1+i, 0, 0); //1+2*(lights.size()+1)+i
				lighting.setValue1i("albedoAtlas["+i+"]", i);
			}
			
			// normal maps // Interesting result if using normals instead of light...
			for (int i = 0; i < lightsNum; i++) {
				GL13.glActiveTexture(GL13.GL_TEXTURE0 + lightsNum + i);
				fbos.bind_texture(1+i, 1, 0); //1+2*(lights.size()+1)+i
				lighting.setValue1i("normalAtlas["+i+"]", lightsNum + i);
			}
			
			// light intensity maps // Interesting result if using normals instead of light...
			for (int i = 0; i < lightsNum; i++) {
				GL13.glActiveTexture(GL13.GL_TEXTURE0 + 2*lightsNum + i);
				fbos.bind_texture(1+i, 2, 0); //1+2*(lights.size()+1)+i
				lighting.setValue1i("lightAtlas["+i+"]", 2*lightsNum + i);
			}
			
			// depthMaps // Interesting result if using normals instead of light...
			for (int i = 0; i < lightsNum; i++) {
				GL13.glActiveTexture(GL13.GL_TEXTURE0 + 3*lightsNum + i);
				fbos.bind_texture(1+i, 3, 0); //1+2*(lights.size()+1)+i
				lighting.setValue1i("depthAtlas["+i+"]", 3*lightsNum + i);
			}
			
			GL13.glActiveTexture(GL13.GL_TEXTURE0);
			
			//camera
			lighting.setValueMat4f("viewMatrix", BufferConverter.toBuffer(scene.getCamera().getViewMatrix()));
			lighting.setValueMat4f("projMatrix", BufferConverter.toBuffer(scene.getCamera().getProjMatrix()));
			
			lighting.setValueMat4f("invViewMatrix", BufferConverter.toBuffer(scene.getCamera().getCameraMatrix()));
			lighting.setValueMat4f("invProjMatrix", BufferConverter.toBuffer(scene.getCamera().getInvProjMatrix()));
			
			for (int i = 0; i < scene.getLights().size(); i++) {
				lighting.setValueMat4f("lightView["+i+"]", BufferConverter.toBuffer(scene.getLights().get(i).getViewMatrix()));
			}
			
			lighting.setValueMat4f("lightProj", BufferConverter.toBuffer(scene.getLights().get(0).getProjMatrix()));
			//lighting.setValue1i("albedoAtlas", 0);
			//lighting.setValue1i("normalAtlas", 1);
			//lighting.setValue1i("depthAtlas", 2);
			//lighting.setValue1i("lightAtlas", 3);
			lighting.setValueVec2f("gridDimension", Scene.GRIDDIMENSION[0], Scene.GRIDDIMENSION[1]);
			
			// draw
			for (int i = 0; i < screenQuad.size(); i++) {
				if (i == 1 && screenQuad.get(i).vertexDataIsUploaded()) {
					screenQuad.get(i).draw(lighting, vbo, textures);
				}
			}
						
			lighting.unbind();
			
			fbos.unbind();
			// ---------------- DRAW END ----------------
			
			// ---------------- DRAW LIGHT ----------------
			
			int index = 0;
			fbos.bindAll(2 +index, 4); // light scene //1 + 4*lights.size() + 1 +index
			fbos.clear(2 +index, 4);
			
			lightingLights.bind();
			
			//fbos.bind_texture(1 + index, 0);
			fbos.bind_texture(0, 0, 0);
			
			GL13.glActiveTexture(GL13.GL_TEXTURE1);
			//fbos.bind_texture(1 + index + lights.size(), 0);
			fbos.bind_texture(0, 0, 1);
			
			GL13.glActiveTexture(GL13.GL_TEXTURE2);
			//fbos.bind_texture(1 + index + 3*lights.size(), 0);
			fbos.bind_texture(0, 0, 3);
			
			// lightAtlas // Interesting result if using normals instead of light...
			for (int i = 0; i < lights.size(); i++) {
				GL13.glActiveTexture(GL13.GL_TEXTURE3+i);
				fbos.bind_texture(1+i, 2, 0); //+2*lights.size()+i, 0
				lightingLights.setValue1i("lightAtlas["+i+"]", 3+i);
			}
			
			GL13.glActiveTexture(GL13.GL_TEXTURE0);
			
			//camera
			lightingLights.setValueMat4f("viewMatrix", BufferConverter.toBuffer(scene.getLights().get(index).getViewMatrix()));
			lightingLights.setValueMat4f("projMatrix", BufferConverter.toBuffer(scene.getLights().get(0).getProjMatrix()));
			
			lightingLights.setValueMat4f("invViewMatrix", BufferConverter.toBuffer(scene.getLights().get(index).getCameraMatrix()));
			lightingLights.setValueMat4f("invProjMatrix", BufferConverter.toBuffer(scene.getLights().get(index).getInvProjMatrix()));
			
			for (int i = 0; i < scene.getLights().size(); i++) {
				lightingLights.setValueMat4f("lightView["+i+"]", BufferConverter.toBuffer(scene.getLights().get(i).getViewMatrix()));
			}
			
			lightingLights.setValueMat4f("lightProj", BufferConverter.toBuffer(scene.getLights().get(0).getProjMatrix()));
			lightingLights.setValue1i("albedoAtlas", 0);
			lightingLights.setValue1i("normalAtlas", 1);
			lightingLights.setValue1i("depthAtlas", 2);
			//lightingLights.setValue1i("lightAtlas", 3);
			lightingLights.setValueVec2f("gridDimension", Scene.GRIDDIMENSION[0], Scene.GRIDDIMENSION[1]);
			
			// draw
			for (int i = 0; i < screenQuad.size(); i++) {
				if (i == 1 && screenQuad.get(i).vertexDataIsUploaded()) {
					screenQuad.get(i).draw(lighting, vbo, textures);
				}
			}
						
			lightingLights.unbind();
			
			fbos.unbind();
			// ---------------- DRAW END ----------------
		} else {
			
			if (objectsWereDrawn) { // wait for camera final
				counter++; // at 10 different..
			}
			
			if (!finishedCPU && objectsWereDrawn && counter>0) { // upload when objects are load
				
				fbos.unbind();
				fbos.unbind_texture();
				
				// upload
				
				/**
				 * FBO order
				 * 
				 * 0 - Original Atlas Image
				 * 
				 * 
				 * 
				 */
				
				// camera
				float[] albedo = 	fbos.getPixelData(1, 0);
				float[] normal = 	fbos.getPixelData(1, 1);
				float[] intensity = fbos.getPixelData(1, 2);
				float[] depth = 	fbos.getPixelData(1, 3);
				int[] size = fbos.getSize(1, 0);
				cpuIllum.uploadViewport(
						size[0], size[1], 
						albedo, normal, depth, intensity,
						BufferConverter.toBuffer(scene.getCamera().getCameraMatrix()),
						BufferConverter.toBuffer(scene.getCamera().getInvProjMatrix()));
				
				// lights
				for (int i = 0; i < lights.size(); i++) {
					albedo = 	fbos.getPixelData(2+i, 0);
					normal = 	fbos.getPixelData(2+i, 1);
					intensity = fbos.getPixelData(2+i, 2);
					depth = 	fbos.getPixelData(2+i, 3);
					float[] depth2 = 	fbos.getPixelData(2+i, 5);
					size = 		fbos.getSize(1+i, 0);
					cpuIllum.uploadLight(
							size[0], size[1], 
							albedo, normal, depth, intensity, depth2,
							BufferConverter.toBuffer(scene.getLights().get(i).getViewMatrix()), 
							BufferConverter.toBuffer(scene.getLights().get(i).getProjMatrix()));
				}
				
				
				float[] result = cpuIllum.calcIllumination();
				
				cpuResultIndex = textures.uploadArray(size[0], size[1], result);
				
				finishedCPU = true;
				
			}
			
		}
		
		// ---------------- DRAW SHADOW ----------------
		drawShadow();
		// ---------------- DRAW END ----------------
		
		// ---------------- FINAL DRAW ----------------
		
		GL11.glClear(GL11.GL_COLOR_BUFFER_BIT | GL11.GL_DEPTH_BUFFER_BIT);
		GL11.glViewport(0, 0, (int)(WindowWrapper.WINDOWSCALE*WindowWrapper.WINDOWX), (int)(WindowWrapper.WINDOWSCALE*WindowWrapper.WINDOWY));
		drawDebug(screenQuad);
		GL11.glViewport(0, 0, WindowWrapper.WINDOWX, WindowWrapper.WINDOWY);
		
		// ---------------- FINAL END ----------------
		
		// drawing to file
		if (timer.waitedEnough() && !drawnToFile) {
			
			ArrayList<ICamera> viewPorts = new ArrayList<ICamera>();
			viewPorts.add(scene.getCamera());
			viewPorts.addAll(scene.getLights());
			
			try {
				ViewportXMLSave.saveObjectDataToXML(viewPorts);
			} catch (Exception e) {
				System.out.println(e);
			}
			
			
			fbos.saveToFile(0, 0, "/content/output/");
			
			drawnToFile = true;
			
		}
		
		Display.update();
		
		timer.endMeasure();
		
		//System.out.println(timer.dt());
		
	}
	
	private void drawShadow() {
		/*if (DRAWSHADOW) {
		
			// WHAT SHADER TO BIND? (reduce shader binds?)
			shadow.bind();
			
			fbos.bind_texture_depth(0);
			
			//camera
			shadow.setValueMat4f("viewMatrix", BufferConverter.toBuffer(scene.getCamera().getViewMatrix()));
			shadow.setValueMat4f("projMatrix", BufferConverter.toBuffer(scene.getCamera().getProjMatrix()));
			
			for (int i = 0; i < scene.getLights().size(); i++) {
				shadow.setValueMat4f("lightView["+i+"]", BufferConverter.toBuffer(scene.getLights().get(i).getViewMatrix()));
			}
			
			shadow.setValueMat4f("lightProj", BufferConverter.toBuffer(scene.getLights().get(0).getProjMatrix()));
			shadow.setValue1i("shadowAtlas", 0);
			
			// draw
			for (int i = 0; i < objects.size(); i++) {
				if (objects.get(i).vertexDataIsUploaded()) {
					objects.get(i).draw(shadow, vbo, textures);
				}
			}
			
			fbos.unbind_texture();
			
			shadow.unbind();
					
			fbos.unbind();
		
		}*/
	}
	
	private void drawDebug(ArrayList<IRenderable> screenQuad) {
		
		// ---------------- SCREENQUAD PASS ----------------
		
		if (!useCPU)
			fbos.bind_texture(1, 4, 0); // first GPU result 25 (for 5 lightsources)
			//fbos.bind_texture(8, 0); // 1 lightsource camera depth buffer
			//fbos.bind_texture(1, 1, 0);
		else
			textures.bindTexture(cpuResultIndex);
		//fbos.bind_texture_1(0); //albedo
		//fbos.bind_texture_2(0); //normal
		//fbos.bind_texture_3(0); //light
		//fbos.bind_texture_depth(0); //depth
		
		//fbos.bind_texture(6, 2, 0); // DEBUG
		
		defaultTexturedShader.bind();
		
		defaultTexturedShader.setValue1i("colorMap", 0);
		
		// draw
		for (int i = 0; i < screenQuad.size(); i++) {
			if (i == 0 && screenQuad.get(i).vertexDataIsUploaded()) {
				screenQuad.get(i).draw(defaultTexturedShader, vbo, textures);
			}
		}
		
		defaultTexturedShader.unbind();
		
		fbos.unbind_texture();
		
		// ---------------- SCREENQUAD END ----------------
		
	}

	@Override
	protected void quit() {
		Display.destroy();
		window.shutdown();
	}
	
}
