package modules.render.cpu;

import java.nio.FloatBuffer;
import java.util.ArrayList;

import modules.render.cpu.boundingsphere.DistanceQuadTree;

import org.lwjgl.util.vector.Matrix4f;
import org.lwjgl.util.vector.Vector2f;
import org.lwjgl.util.vector.Vector3f;
import org.lwjgl.util.vector.Vector4f;

/**
 * Problem: no interpolation at lookup
 * 
 * @author DarkMatt
 *
 */

public class CPUIllum {

	// scene 0: 512 px: 0.0007 (too much - aliasing), 128: 0.005
	// scene 1: 128 px: 0.01, 256 px: 0.005, 512px 0.001f - bias not enough!
	private float depth_bias = 0.005f; // 0.007 (64 px), 0.005f (64 px); 0.0035 (128 px - double size SM); 0.001f (420 px); 0.0007f (512px); 0.001f (128 px - double size SM); 0.0005f (840 px) - depends on resoluation!
	// intensity weighting for the first bounce
	// scene 0: 0.3,scene 1: 1.7f, scene 2: 1.2f
	private float direct_factor = 0.3f; //0.2 .3 .13 0.4
	// intensity weighting for the 2nd bounce
	private float indirect_factor = 1.0f;//0.75f;
	
	private boolean depthMultiChannel = false;
	
	private int[] c_size;

	private ImagePlane<Vector4f, Vector3f, Float, Float> cameraPlane;
	private ArrayList<ImagePlane<Vector4f, Vector3f, Float, Float>> lightPlanes;
	private ArrayList<ImagePlanePosition<Vector4f, Vector3f, Float, Float>> lightPlanePositions;
	
	private ArrayList<Vector4f> lightPosition;
	
	// direct illumination result
	private ArrayList<Vector4f[][]> directIllum; // Vector4 because the Vector3 cast costs time
	private ArrayList<Vector4f[][]> directIllumSingle; // only direct illum from one lightsource
	
	private Vector4f[][] result;
	
	// measuring time
	private long startTime = 0;
	
	public CPUIllum() {
		lightPlanes = new ArrayList<ImagePlane<Vector4f, Vector3f, Float, Float>>();
		lightPlanePositions = new ArrayList<ImagePlanePosition<Vector4f, Vector3f, Float, Float>>();
		lightPosition = new ArrayList<Vector4f>();
		
		directIllum = new ArrayList<Vector4f[][]>();
		directIllumSingle = new ArrayList<Vector4f[][]>();
	}
	
	/**
	 * Linearizes a depth buffer. (This only works if 32 bit textures are used for enough precision. 8 bit does not suffice.)
	 * 
	 * @param depth
	 * @return
	 */
	private float[] linearize(float[] depth) {
		float n = 0.1f; // camera z near
		float f = 100.0f; // camera z far
		
		float[] out = new float[depth.length];
		for (int i = 0; i < depth.length; i++) {
			out[i] = (float)(2.0 * n) / (f + n - depth[i] * (f - n));
		}
		
		//System.out.println("Did linearize.");
		
		return out;
	}
	
	/**
	 * Arrays are assumed to be filled with 4 values for each array (RGBA), regardless of what information they contain.
	 * (Alternatively the offset while reading would have been changed; 4*index versus 3*index ...)
	 * 
	 * @param sx
	 * @param sy
	 * @param albedo
	 * @param normal
	 * @param depth
	 * @param light
	 */
	public void uploadViewport(int sx, int sy, float[] albedo, float[] normal, float[] depth, float[] light, FloatBuffer inverseViewMatrix, FloatBuffer inverseProjMatrix) {
		c_size = new int[]{sx, sy};
		
		//albedo = linearize(albedo); // linearize for displaying depth
		
		Matrix4f invViewMatrix = (Matrix4f)new Matrix4f().loadTranspose(inverseViewMatrix); //!!! transpose needed
		Matrix4f invProjMatrix = (Matrix4f)new Matrix4f().loadTranspose(inverseProjMatrix);
		
		result = new Vector4f[sx][sy];
		
		cameraPlane = new ImagePlane<Vector4f, Vector3f, Float, Float>(sx, sy);
		cameraPlane.setInvViewMatrix(invViewMatrix);
		cameraPlane.setInvProjMatrix(invProjMatrix);
		
		int index = 0;
		for (int y = 0; y < sy; y++) {
			for (int x = 0; x < sx; x++) {
				cameraPlane.setAlbedo(x, y, new Vector4f(albedo[4*index], albedo[4*index+1], albedo[4*index+2], albedo[4*index+3]));
				
				Vector3f n = new Vector3f(normal[4*index], normal[4*index+1], normal[4*index+2]);
				VectorMath.toNegativePositive(n); // from 0..1 to -1..1
				
				cameraPlane.setNormal(x, y, n);
				if (depthMultiChannel) {

					int r = (int)(255.0f*depth[4*index]);
					int g = (int)(255.0f*depth[4*index+1]);
					int b = (int)(255.0f*depth[4*index+2]);
					
					int dint = (r >> 16) | (g >> 8) | b;
					
					float d = (float)dint/128.0f;
					
					cameraPlane.setDepth(x, y, d);
				} else {
					cameraPlane.setDepth(x, y, depth[4*index]);
				}
				cameraPlane.setLight(x, y, light[4*index]);
				
				index ++;
			}
		}
	}
	
	/**
	 * Depth 2 is a higher resolution depth map for shadow-mapping
	 * 
	 * @param sx
	 * @param sy
	 * @param albedo
	 * @param normal
	 * @param depth
	 * @param light
	 * @param depth2
	 * @param lightViewMatrix
	 * @param lightProjMatrix
	 */
	public void uploadLight(int sx, int sy, float[] albedo, float[] normal, float[] depth, float[] light, float[] depth2, FloatBuffer lightViewMatrix, FloatBuffer lightProjMatrix) {
		
		ImagePlane<Vector4f, Vector3f, Float, Float> image = new ImagePlane<Vector4f, Vector3f, Float, Float>(sx, sy);
		
		int index = 0;
		for (int y = 0; y < sy; y++) {
			for (int x = 0; x < sx; x++) {
				image.setAlbedo(x, y, new Vector4f(albedo[4*index], albedo[4*index+1], albedo[4*index+2], albedo[4*index+3]));
				
				Vector3f n = new Vector3f(normal[4*index], normal[4*index+1], normal[4*index+2]);
				VectorMath.toNegativePositive(n);
				
				image.setNormal(x, y, n);
				if (depthMultiChannel) { // just for loading images files
					
					int r = (int)(255.0f*depth[4*index]);
					int g = (int)(255.0f*depth[4*index+1]);
					int b = (int)(255.0f*depth[4*index+2]);
					
					int dint = (r >> 16) | (g >> 8) | b;
					
					float d = (float)dint/128.0f;
					
					image.setDepth(x, y, d);
				} else {
					image.setDepth(x, y, depth[4*index]);
				}
				image.setLight(x, y, light[4*index]);
				
				index ++;
			}
		}
		
		// fill depth2
		index = 0;
		for (int y = 0; y < 2*sy; y++) {
			for (int x = 0; x < 2*sx; x++) {
				if (depthMultiChannel) { // just for loading images files
					
					int r = (int)(255.0f*depth2[4*index]);
					int g = (int)(255.0f*depth2[4*index+1]);
					int b = (int)(255.0f*depth2[4*index+2]);
					
					int dint = (r >> 16) | (g >> 8) | b;
					
					float d = (float)dint/128.0f;
					
					image.setDepth2(x, y, d);
				} else {
					image.setDepth2(x, y, depth2[4*index]);
				}
				
				index ++;
			}
		}
		
		Matrix4f lvm = new Matrix4f();
		lvm.loadTranspose(lightViewMatrix);
		image.setViewMatrix(lvm);
		
		Matrix4f lpm = new Matrix4f();
		lpm.loadTranspose(lightProjMatrix);
		image.setProjMatrix(lpm);
		
		Vector4f lpos = new Vector4f(0,0,0,1);
		Matrix4f inv = new Matrix4f(image.getInvViewMatrix());
		Matrix4f.transform(inv, lpos, lpos);
		//inv.invert();
		//Vector4f lpos = new Vector4f(inv.m03, inv.m13, inv.m23, inv.m33);
		/*lpos.x /= lpos.w;
		lpos.y /= lpos.w;
		lpos.z /= lpos.w;
		lpos.w = 1.0f;*/
		lightPosition.add(lpos); // -2.470907E-9 -0.0 -0.026666671 1.0 // -6.7132845 25.000002 38.07292 1.0
		
		lightPlanes.add(image);
	}
	
	private Vector4f getWorldSpacePosition(Vector2f textCoord) {

		Vector2f relCoord = new Vector2f((float)textCoord.x/(float)(c_size[0]-1), (float)textCoord.y/(float)(c_size[1]-1));
		
		int[] coord = new int[]{(int)textCoord.x, (int)textCoord.y};//new int[]{(int)(textCoord.x*(c_size[0]-1)), ((int)textCoord.y*(c_size[1]-1))};
		
		float depth = cameraPlane.Depth(coord[0], coord[1]);//texture2D(depthAtlas[0], textCoord).x;
		Vector4f screenSpacePos = new Vector4f(relCoord.x, relCoord.y, depth, 1.0f);
	    screenSpacePos.x = 2.0f*(screenSpacePos.x-0.5f);
	    screenSpacePos.y = 2.0f*(screenSpacePos.y-0.5f);
	    screenSpacePos.z = 2.0f*(screenSpacePos.z-0.5f);
	    screenSpacePos.w = 2.0f*(screenSpacePos.w-0.5f);
	    Vector4f worldSpace = new Vector4f(screenSpacePos);
	    Matrix4f.transform(cameraPlane.getInvProjMatrix(), screenSpacePos, worldSpace); //invProjMatrix
	    Matrix4f.transform(cameraPlane.getInvViewMatrix(), worldSpace, worldSpace); //invViewMatrix
	    worldSpace.x /= worldSpace.w;
	    worldSpace.y /= worldSpace.w;
	    worldSpace.z /= worldSpace.w;
	    worldSpace.w = 1.0f;
	    //Vector4f worldSpace = invViewMatrix * invProjMatrix * screenSpacePos;
	    //worldSpace / worldSpace.w;
	    return worldSpace;
	}
	
	/**
	 * Already divided by w.
	 * 
	 * @param textCoord
	 * @param index
	 * @return
	 */
	private Vector4f getWorldSpacePositionLightsource(Vector2f textCoord, int index) {

		Vector2f relCoord = new Vector2f((float)textCoord.x/(float)(c_size[0]-1), (float)textCoord.y/(float)(c_size[1]-1));
		
		int[] coord = new int[]{(int)textCoord.x, (int)textCoord.y};//new int[]{(int)(textCoord.x*(c_size[0]-1)), ((int)textCoord.y*(c_size[1]-1))};
		
		float depth = lightPlanes.get(index).Depth(coord[0], coord[1]);//texture2D(depthAtlas[0], textCoord).x;
		Vector4f screenSpacePos = new Vector4f(relCoord.x, relCoord.y, depth, 1.0f);
	    screenSpacePos.x = 2.0f*(screenSpacePos.x-0.5f);
	    screenSpacePos.y = 2.0f*(screenSpacePos.y-0.5f);
	    screenSpacePos.z = 2.0f*(screenSpacePos.z-0.5f);
	    screenSpacePos.w = 2.0f*(screenSpacePos.w-0.5f);
	    Vector4f worldSpace = new Vector4f(screenSpacePos);
	    Matrix4f.transform(lightPlanes.get(index).getInvProjMatrix(), screenSpacePos, worldSpace); //invProjMatrix
	    Matrix4f.transform(lightPlanes.get(index).getInvViewMatrix(), worldSpace, worldSpace); //invViewMatrix
	    worldSpace.x /= worldSpace.w;
	    worldSpace.y /= worldSpace.w;
	    worldSpace.z /= worldSpace.w;
	    worldSpace.w = 1.0f;
	    //Vector4f worldSpace = invViewMatrix * invProjMatrix * screenSpacePos;
	    //worldSpace / worldSpace.w;
	    return worldSpace;
	}
	
	private Vector4f perPixel(Vector2f coord) {
		Vector4f out;
		
		//vec4 ws_Position = getWorldSpacePosition(pass_TextCoord);
		out = getWorldSpacePosition(coord);
		
		return out;
	}
	
	private Vector4f perPixelLightsource(Vector2f coord, int index) {
		Vector4f out;
		
		//vec4 ws_Position = getWorldSpacePosition(pass_TextCoord);
		out = getWorldSpacePositionLightsource(coord, index);
		
		return out;
	}
	
	/**
	 * Calculates the direct illumination image for the image plane.
	 * 
	 * @return
	 */
	private Vector4f[][] calcDirect() {
		
		Vector4f[][] result = new Vector4f[c_size[0]][c_size[1]];
		
		// for all pixels
		for (int y = 0; y < c_size[1]; y++) {
			for (int x = 0; x < c_size[0]; x++) {
				
				result[x][y] = new Vector4f(cameraPlane.Albedo(x, y)); // needs a COPY!
				
				Vector4f ws_Position = this.perPixel(new Vector2f(x,y));
				
				Vector3f normal = new Vector3f(cameraPlane.Normal(x, y));
				/*normal.x = 2.0f*(normal.x-0.5f); // -1 to 1 //NECESSARY?
				normal.y = 2.0f*(normal.y-0.5f);
				normal.z = 2.0f*(normal.z-0.5f);
				normal.normalise();*/
				
				Vector3f lightColor = new Vector3f(0.0f, 0.0f, 0.0f);
				
				// over all light-sources
				for (int i = 0; i < lightPlanes.size(); i++) { //l_albedo.size()
					
					Vector4f lightOrigin = lightPosition.get(i);
					Vector3f lightDir = new Vector3f();
					lightDir.x = lightOrigin.x - ws_Position.x;
					lightDir.y = lightOrigin.y - ws_Position.y;
					lightDir.z = lightOrigin.z - ws_Position.z;
					lightDir.normalise();
					
					/*
					vec4 lightPos = lightMatrix[i] * ws_Position; // in light-space
					lightPos /= lightPos.w; // -1 .. 1
					
					lightPos /= 2.0; // to 0 .. 1
					lightPos += 0.5;
					 */
					
					Vector4f lightPos = new Vector4f(); // in light-space
					//Matrix4f.transform(lightMatrices.get(i), ws_Position, lightPos);
					Matrix4f.transform(lightPlanes.get(i).getViewProjMatrix(), ws_Position, lightPos);
					lightPos.x /= lightPos.w; // -1 .. 1
					lightPos.y /= lightPos.w;
					lightPos.z /= lightPos.w;
					lightPos.w = 1.0f;
					
					lightPos.x = 0.5f*(lightPos.x+1.0f); // to 0 .. 1
					lightPos.y = 0.5f*(lightPos.y+1.0f);
					lightPos.z = 0.5f*(lightPos.z+1.0f);
					lightPos.w = 0.5f*(lightPos.w+1.0f);
					
					//float factor = 0.3f;
					
					int[] coord = new int[]{
							(int)(lightPos.x*(c_size[0]-1)), (int)(lightPos.y*(c_size[1]-1))
					};
					float depth = 1.0f;
					if (coord[0] > 0 && coord[0] < c_size[0] && coord[1] > 0 && coord[1] < c_size[1]) {
						//depth = l_depth.get(i)[coord[0]][coord[1]];
						depth = lightPlanes.get(i).Depth(coord[0], coord[1]);
					}
					
					if (lightPos.z > depth + depth_bias) { // + bias // COMPARE camera space pixel position TO camera depth map
						//lightColor += vec3(0.0, 0.0, 0.0);
					} else {
						//vec4 light = textureLod(lightAtlas[i], lightPos.xy, 0);
						float light = 0.0f;
						if (coord[0] > 0 && coord[0] < c_size[0] && coord[1] > 0 && coord[1] < c_size[1]) {
							//light = l_light.get(i)[coord[0]][coord[1]];
							light = lightPlanes.get(i).Light(coord[0], coord[1]);
						}
						
						float ndotl = Vector3f.dot(lightDir, normal);
						
						if (ndotl < 0) {
							ndotl = 0;
						}
						if (ndotl > 1) {
							ndotl = 1;
						}
						
						lightColor.x += direct_factor * ndotl * light;
						lightColor.y += direct_factor * ndotl * light;
						lightColor.z += direct_factor * ndotl * light;
					}
					
				}
				
				
				result[x][y].x *= lightColor.x;
				result[x][y].y *= lightColor.y;
				result[x][y].z *= lightColor.z;
				
				//result[x][y] = worldspacePosition;
				//result[x][y] = new Vector4f(normal.x, normal.y, normal.z, 1.0f);
				//result[x][y] = new Vector4f(c_depth[x][y],c_depth[x][y],c_depth[x][y],1.0f);
				
			}
		}
		
		return result;
		
	}
	
	private Vector4f[][] calcDirectLightSource(int index) {
		
		Vector4f[][] result = new Vector4f[c_size[0]][c_size[1]];
		
		// for all pixels
		for (int y = 0; y < c_size[1]; y++) {
			for (int x = 0; x < c_size[0]; x++) {
				
				result[x][y] = new Vector4f(lightPlanes.get(index).Albedo(x, y)); // needs a COPY!
				
				Vector4f ws_Position = this.perPixelLightsource(new Vector2f(x,y), index);
				
				Vector3f normal = new Vector3f(lightPlanes.get(index).Normal(x, y));
				/*normal.x = 2.0f*(normal.x-0.5f); // -1 to 1
				normal.y = 2.0f*(normal.y-0.5f);
				normal.z = 2.0f*(normal.z-0.5f);
				normal.normalise();*/
				
				//float factor = 0.3f;
				
				// calc light from camera perspective
				
				Vector4f lightOrigin = lightPosition.get(index);
				Vector3f lightDir = new Vector3f();
				lightDir.x = lightOrigin.x - ws_Position.x;
				lightDir.y = lightOrigin.y - ws_Position.y;
				lightDir.z = lightOrigin.z - ws_Position.z;
				lightDir.normalise();
				
				float ndotl = Vector3f.dot(lightDir, normal);
				
				if (ndotl < 0) {
					ndotl = 0;
				}
				if (ndotl > 1) {
					ndotl = 1;
				}
				
				float lightIntensity = lightPlanes.get(index).Light(x, y);
				
				// calc light from camera perspective end
				
				Vector3f lightColor = new Vector3f(direct_factor * ndotl * lightIntensity, direct_factor * ndotl * lightIntensity, direct_factor * ndotl * lightIntensity);
				
				// over all light-sources
				for (int i = 0; i < lightPlanes.size(); i++) { //l_albedo.size()
					
					if ( i != index) {
					
						lightOrigin = lightPosition.get(i);
						lightDir = new Vector3f();
						lightDir.x = lightOrigin.x - ws_Position.x;
						lightDir.y = lightOrigin.y - ws_Position.y;
						lightDir.z = lightOrigin.z - ws_Position.z;
						lightDir.normalise();
						
						/*
						vec4 lightPos = lightMatrix[i] * ws_Position; // in light-space
						lightPos /= lightPos.w; // -1 .. 1
						
						lightPos /= 2.0; // to 0 .. 1
						lightPos += 0.5;
						 */
						
						Vector4f lightPos = new Vector4f(); // in light-space
						//Matrix4f.transform(lightMatrices.get(i), ws_Position, lightPos);
						Matrix4f.transform(lightPlanes.get(i).getViewProjMatrix(), ws_Position, lightPos);
						lightPos.x /= lightPos.w; // -1 .. 1
						lightPos.y /= lightPos.w;
						lightPos.z /= lightPos.w;
						lightPos.w = 1.0f;
						
						lightPos.x = 0.5f*(lightPos.x+1.0f); // to 0 .. 1
						lightPos.y = 0.5f*(lightPos.y+1.0f);
						lightPos.z = 0.5f*(lightPos.z+1.0f);
						lightPos.w = 0.5f*(lightPos.w+1.0f);
						
						int[] coord = new int[]{
								(int)(lightPos.x*(c_size[0]-1)), (int)(lightPos.y*(c_size[1]-1))
						};
						float depth = 1.0f;
						if (coord[0] > 0 && coord[0] < c_size[0] && coord[1] > 0 && coord[1] < c_size[1]) {
							//depth = l_depth.get(i)[coord[0]][coord[1]];
							depth = lightPlanes.get(i).Depth(coord[0], coord[1]);
						}
						
						if (lightPos.z > depth + depth_bias) { // + bias // COMPARE camera space pixel position TO camera depth map
							//lightColor += vec3(0.0, 0.0, 0.0);
						} else {
							//vec4 light = textureLod(lightAtlas[i], lightPos.xy, 0);
							float light = 0.0f;
							if (coord[0] > 0 && coord[0] < c_size[0] && coord[1] > 0 && coord[1] < c_size[1]) {
								//light = l_light.get(i)[coord[0]][coord[1]];
								light = lightPlanes.get(i).Light(coord[0], coord[1]);
							}
							
							ndotl = Vector3f.dot(lightDir, normal);
							
							if (ndotl < 0) {
								ndotl = 0;
							}
							if (ndotl > 1) {
								ndotl = 1;
							}
							
							lightColor.x += direct_factor * ndotl * light;
							lightColor.y += direct_factor * ndotl * light;
							lightColor.z += direct_factor * ndotl * light;
						}
						
					}
					
				}
				
				
				result[x][y].x *= lightColor.x;
				result[x][y].y *= lightColor.y;
				result[x][y].z *= lightColor.z;
				
				//result[x][y] = worldspacePosition;
				//result[x][y] = new Vector4f(normal.x, normal.y, normal.z, 1.0f);
				//result[x][y] = new Vector4f(c_depth[x][y],c_depth[x][y],c_depth[x][y],1.0f);
				
			}
		}
		
		return result;
		
	}
	
	/**
	 * Only for the current light-source.
	 * 
	 * @param index
	 * @return
	 */
	private Vector4f[][] calcDirectLightSourceSingle(int index) {
		
		Vector4f[][] result = new Vector4f[c_size[0]][c_size[1]];
		
		// for all pixels
		for (int y = 0; y < c_size[1]; y++) {
			for (int x = 0; x < c_size[0]; x++) {
				
				result[x][y] = new Vector4f(lightPlanes.get(index).Albedo(x, y)); // needs a COPY!
				
				Vector4f ws_Position = this.perPixelLightsource(new Vector2f(x,y), index);
				
				Vector3f normal = new Vector3f(lightPlanes.get(index).Normal(x, y));
				/*normal.x = 2.0f*(normal.x-0.5f); // -1 to 1
				normal.y = 2.0f*(normal.y-0.5f);
				normal.z = 2.0f*(normal.z-0.5f);
				normal.normalise();*/
				
				//float factor = 0.3f;
				
				// calc light from camera perspective
				
				Vector4f lightOrigin = lightPosition.get(index);
				Vector3f lightDir = new Vector3f();
				lightDir.x = lightOrigin.x - ws_Position.x; // vector from pixel to light
				lightDir.y = lightOrigin.y - ws_Position.y;
				lightDir.z = lightOrigin.z - ws_Position.z;
				lightDir.normalise();
				
				float ndotl = Vector3f.dot(lightDir, normal);
				
				if (ndotl < 0) {
					ndotl = 0;
				}
				if (ndotl > 1) {
					ndotl = 1;
				}
				
				float lightIntensity = lightPlanes.get(index).Light(x, y);
				
				// calc light from camera perspective end
				
				//ndotl = 1;
				
				Vector3f lightColor = new Vector3f(direct_factor * ndotl * lightIntensity, direct_factor * ndotl * lightIntensity, direct_factor * ndotl * lightIntensity);
				
				result[x][y].x *= lightColor.x;
				result[x][y].y *= lightColor.y;
				result[x][y].z *= lightColor.z;
				
				/*result[x][y].x = ws_Position.x;
				result[x][y].y = ws_Position.y;
				result[x][y].z = ws_Position.z;*/
				
				//result[x][y] = worldspacePosition;
				//result[x][y] = new Vector4f(normal.x, normal.y, normal.z, 1.0f);
				//result[x][y] = new Vector4f(c_depth[x][y],c_depth[x][y],c_depth[x][y],1.0f);
				
			}
		}
		
		return result;
		
	}
	
	/**
	 * 
	 * Version A) sums over the summed direct illumination
	 * 
	 * (not renewed)
	 * 
	 * first solution
	 * 
	 * @return
	 */
	private Vector4f[][] calcIndirectIllumSummed() {
		
		Vector4f[][] result = new Vector4f[c_size[0]][c_size[1]];
		
		// for all pixels
		for (int y = 0; y < c_size[1]; y++) {
			for (int x = 0; x < c_size[0]; x++) {
				
				double[] newLight = new double[]{0, 0, 0};
				
				//Vector4f newLight = new Vector4f(directIllum.get(0)[x][y]); // direct illumination
				
				Vector4f pixelColor = new Vector4f(cameraPlane.Albedo(x, y));
				Vector4f pixelPos = getWorldSpacePosition(new Vector2f(x, y));
				
				// indirect
				// over CAMERA
				
				// over all pixels
				for (int ly = 0; ly < c_size[1]; ly++) {
					for (int lx = 0; lx < c_size[0]; lx++) {
						
						if (x != lx && y != ly) {
						
							double factor = 1.0;
							
							Vector4f lightPixelColor = new Vector4f(directIllum.get(0)[lx][ly]);
							Vector4f lightPixelPos = getWorldSpacePosition(new Vector2f(lx, ly));
							Vector3f lightPixelNormal = new Vector3f(cameraPlane.Normal(lx, ly));
							
							Vector4f lightOrigin = lightPixelPos; //WRONG: lightDir = pixel to pixel direction!
							Vector3f lightDir = new Vector3f();
							lightDir.x = pixelPos.x - lightOrigin.x;
							lightDir.y = pixelPos.y - lightOrigin.y;
							lightDir.z = pixelPos.z - lightOrigin.z;
							lightDir.normalise();
							
							// from g-buffer frag
							//float d = length(pass_LightSpacePosition.xyz)/45;
							//d = 1.0/pow(d,2.0);
							
							float ndotl = Vector3f.dot(lightDir, lightPixelNormal);
							if (ndotl < 0) {
								ndotl = 0;
							}
							if (ndotl > 1) {
								ndotl = 1;
							}
							
							Vector3f pPos = new Vector3f(pixelPos.x/pixelPos.w, pixelPos.y/pixelPos.w, pixelPos.z/pixelPos.w);
							Vector3f lPPos = new Vector3f(lightPixelPos.x/lightPixelPos.w, lightPixelPos.y/lightPixelPos.w, lightPixelPos.z/lightPixelPos.w);
							Vector3f dst = new Vector3f();
							Vector3f.sub(pPos, lPPos, dst);
							double len = (double)dst.length()*1.0; //4m 100 too bright..
							
							// see http://imdoingitwrong.wordpress.com/2011/01/31/light-attenuation/
							// http://tomdalling.com/blog/modern-opengl/07-more-lighting-ambient-specular-attenuation-gamma/
							double attenuation = 1.0/(1.0f + len*len);
							
							// surface color times light
							newLight[0] += (factor * attenuation * ndotl * lightPixelColor.x * pixelColor.x);
							newLight[1] += (factor * attenuation * ndotl * lightPixelColor.y * pixelColor.y);
							newLight[2] += (factor * attenuation * ndotl * lightPixelColor.z * pixelColor.z);
							
							/*newLight.x += (factor * d * lightPixelColor.x);
							newLight.y += (factor * d * lightPixelColor.y);
							newLight.z += (factor * d * lightPixelColor.z);
							newLight.w += (factor * d * lightPixelColor.w);*/
						
						} // end condition
							
					}
				}
				
				// over all LIGHT_SOURCES
				/*for (int i = 0; i < 1; i++) {
					
					Vector4f lightOrigin = lightPosition.get(i); //WRONG: lightDir = pixel to pixel direction!
					Vector3f lightDir = new Vector3f();
					lightDir.x = lightOrigin.x - pixelPos.x;
					lightDir.y = lightOrigin.y - pixelPos.y;
					lightDir.z = lightOrigin.z - pixelPos.z;
					lightDir.normalise();
					
					// over all pixels
					for (int ly = 0; ly < c_size[1]; ly++) {
						for (int lx = 0; lx < c_size[0]; lx++) {
							
							double factor = 1.0;
							
							Vector4f lightPixelColor = new Vector4f(directIllum.get(i+1)[lx][ly]);
							Vector4f lightPixelPos = getWorldSpacePositionLightsource(new Vector2f(lx, ly), i);
							Vector3f lightPixelNormal = new Vector3f(lightPlanes.get(i).Normal(lx, ly));
							
							// from g-buffer frag
							//float d = length(pass_LightSpacePosition.xyz)/45;
							//d = 1.0/pow(d,2.0);
							
							float ndotl = Vector3f.dot(lightDir, lightPixelNormal);
							if (ndotl < 0) {
								ndotl = 0;
							}
							if (ndotl > 1) {
								ndotl = 1;
							}
							
							Vector3f pPos = new Vector3f(pixelPos.x/pixelPos.w, pixelPos.y/pixelPos.w, pixelPos.z/pixelPos.w);
							Vector3f lPPos = new Vector3f(lightPixelPos.x/lightPixelPos.w, lightPixelPos.y/lightPixelPos.w, lightPixelPos.z/lightPixelPos.w);
							Vector3f dst = new Vector3f();
							Vector3f.sub(pPos, lPPos, dst);
							double len = (double)dst.length()*4.0; //100 too bright..
							double d = 0.0;
							if (len < 100.0) {
								d = 1.0/(len*len);
							}
							if (len < 1) { // numerics problem?
								d = 1;
							}
							
							// surface color times light
							newLight[0] += (factor * d * ndotl * lightPixelColor.x * pixelColor.x);
							newLight[1] += (factor * d * ndotl * lightPixelColor.y * pixelColor.y);
							newLight[2] += (factor * d * ndotl * lightPixelColor.z * pixelColor.z);
							
							/*newLight.x += (factor * d * lightPixelColor.x);
							newLight.y += (factor * d * lightPixelColor.y);
							newLight.z += (factor * d * lightPixelColor.z);
							newLight.w += (factor * d * lightPixelColor.w);*/
							
						/*}
					}
					
				}*/
				
				double size = c_size[0] * c_size[1]; // for normalization
				
				size /= 100.0; //500
				
				newLight[0] /= size;
				newLight[1] /= size;
				newLight[2] /= size;
				
				double[] light = new double[]{
						directIllum.get(0)[x][y].x + newLight[0], 
						directIllum.get(0)[x][y].y + newLight[1], 
						directIllum.get(0)[x][y].z + newLight[2]};
				
				//if (light[0] > 0.1) {
				//	result[x][y] = new Vector4f(1.0f, 0.0f, 0.0f, 1.0f);
				//} else {
					result[x][y] = new Vector4f((float)light[0], (float)light[1], (float)light[2], directIllum.get(0)[x][y].w);
				//}
				
				//directIllum.get(0)[x][y].x, directIllum.get(0)[x][y].y, directIllum.get(0)[x][y].z
				
			}
			
			System.out.println("row " + y);
			
		}
		
		return result;
	}
	
	/**
	 * 
	 * Version B) sums only over single input sources
	 * 
	 * @return
	 */
	private Vector4f[][] calcIndirectIllumSingle() {
		
		Vector4f[][] result = new Vector4f[c_size[0]][c_size[1]];
		
		// for all pixels
		for (int y = 0; y < c_size[1]; y++) {
			for (int x = 0; x < c_size[0]; x++) {
				
				double[] lightSum = new double[]{0, 0, 0};
				
				//Vector4f newLight = new Vector4f(directIllum.get(0)[x][y]); // direct illumination
				
				Vector4f pixelColor = new Vector4f(cameraPlane.Albedo(x, y));
				Vector4f pixelPos = getWorldSpacePosition(new Vector2f(x, y));
				Vector3f pixelNormal = new Vector3f(cameraPlane.Normal(x, y));
				
				// indirect
				// over CAMERA
				
				// over all LIGHT_SOURCES directIllumSingle.size()
				for (int i = 0; i < directIllumSingle.size(); i++) {
					
					double[] newLight = new double[]{0, 0, 0};
					
					// over all pixels
					for (int ly = 0; ly < c_size[1]; ly++) {
						for (int lx = 0; lx < c_size[0]; lx++) {
							
							Vector4f lightPixelColor = new Vector4f(directIllumSingle.get(i)[lx][ly]);
							Vector4f lightPixelPos = getWorldSpacePositionLightsource(new Vector2f(lx, ly), i);
							Vector3f lightPixelNormal = new Vector3f(lightPlanes.get(i).Normal(lx, ly));
							//Vector3f lightPixelNormal = new Vector3f(lightPlanes.get(i).Normal(lx, ly));
							/*lightPixelNormal.x = 2.0f*(lightPixelNormal.x-0.5f); // -1 to 1
							lightPixelNormal.y = 2.0f*(lightPixelNormal.y-0.5f);
							lightPixelNormal.z = 2.0f*(lightPixelNormal.z-0.5f);
							lightPixelNormal.normalise();*/
							
							//WRONG: lightDir = pixel to pixel direction!
							Vector3f lightDir = new Vector3f();
							lightDir.x = lightPixelPos.x - pixelPos.x; //swapped; pixelPos ... position of the current pixel
							lightDir.y = lightPixelPos.y - pixelPos.y;
							lightDir.z = lightPixelPos.z - pixelPos.z;
							lightDir.normalise();
							
							// from g-buffer frag
							//float d = length(pass_LightSpacePosition.xyz)/45;
							//d = 1.0/pow(d,2.0);
							
							float ndotl = Vector3f.dot(lightDir, pixelNormal); //lightPixelNormal
							if (ndotl < 0) {
								ndotl = 0;
							}
							if (ndotl > 1) {
								ndotl = 1;
							}
							//ndotl = 1;
							
							// for the light pixel
							float ndotl_light = Vector3f.dot((Vector3f)lightDir.negate(), lightPixelNormal);
							if (ndotl_light < 0) {
								ndotl_light = 0;
							}
							if (ndotl_light > 1) {
								ndotl_light = 1;
							}
							
							Vector3f pPos = new Vector3f(pixelPos.x, pixelPos.y, pixelPos.z);
							Vector3f lPPos = new Vector3f(lightPixelPos.x, lightPixelPos.y, lightPixelPos.z);
							Vector3f dst = new Vector3f();
							Vector3f.sub(pPos, lPPos, dst);
							double len = (double)dst.length()*1.0;
							double attenuation = 1.0/(1.0f + len*len);
							
							// surface color times light
							newLight[0] += (indirect_factor * attenuation * ndotl * ndotl_light * lightPixelColor.x * pixelColor.x);
							newLight[1] += (indirect_factor * attenuation * ndotl * ndotl_light * lightPixelColor.y * pixelColor.y);
							newLight[2] += (indirect_factor * attenuation * ndotl * ndotl_light * lightPixelColor.z * pixelColor.z);
							
							/*newLight[0] += ndotl*d/5.0f;
							newLight[1] += ndotl*d/5.0f;
							newLight[2] += ndotl*d/5.0f;*/
							
							/*newLight[0] += d/15.0f; // LOOKS COOL (distance sum) - maybe artifact at shadow/object boundaries?
							newLight[1] += d/15.0f;
							newLight[2] += d/15.0f;*/
							
						}
					}
					
					double size = c_size[0] * c_size[1]; // for normalization
					
					size /= 100.0; //100, 300, 500
					
					newLight[0] /= size;
					newLight[1] /= size;
					newLight[2] /= size;
					
					lightSum[0] += newLight[0];
					lightSum[1] += newLight[1];
					lightSum[2] += newLight[2];
					
				}
				
				double[] light = new double[]{
						directIllum.get(0)[x][y].x + lightSum[0], 
						directIllum.get(0)[x][y].y + lightSum[1], 
						directIllum.get(0)[x][y].z + lightSum[2]};
				
				/*light = new double[]{		//DEBUG DISPLAY for SINGLOUTPUT
						1.0f*lightSum[0], 
						1.0f*lightSum[1], 
						1.0f*lightSum[2]};*/
				
				//if (light[0] > 0.1) {
				//	result[x][y] = new Vector4f(1.0f, 0.0f, 0.0f, 1.0f);
				//} else {
					result[x][y] = new Vector4f((float)light[0], (float)light[1], (float)light[2], directIllum.get(0)[x][y].w);
				//}
				
				//directIllum.get(0)[x][y].x, directIllum.get(0)[x][y].y, directIllum.get(0)[x][y].z
				
			}
			
			System.out.println("row " + y);
			
		}
		
		return result;
	}
	
	
	// VERSION OPTIMIZED BUT NO QUADTREE
	private Vector4f[][] optimizedCalcIndirectIllumSingleNoQuadtree() {
		
		Vector4f[][] result = new Vector4f[c_size[0]][c_size[1]];
		
		// for all pixels
		for (int y = 0; y < c_size[1]; y++) {
			for (int x = 0; x < c_size[0]; x++) {
				
				double[] lightSum = new double[]{0, 0, 0};
				
				//Vector4f newLight = new Vector4f(directIllum.get(0)[x][y]); // direct illumination
				
				Vector4f pixelColor = new Vector4f(cameraPlane.Albedo(x, y));
				Vector4f pixelPos = getWorldSpacePosition(new Vector2f(x, y));
				Vector3f pixelNormal = new Vector3f(cameraPlane.Normal(x, y));
				
				// indirect
				// over CAMERA
				
				// over all LIGHT_SOURCES directIllumSingle.size()
				for (int i = 0; i < directIllumSingle.size(); i++) {
					
					double[] newLight = new double[]{0, 0, 0};
					
					// over all pixels
					for (int ly = 0; ly < c_size[1]; ly++) {
						for (int lx = 0; lx < c_size[0]; lx++) {
							
							Vector4f lightPixelColor = new Vector4f(directIllumSingle.get(i)[lx][ly]);
							Vector3f lightPixelPos = lightPlanePositions.get(i).getWs_position(lx, ly);//getWorldSpacePositionLightsource(new Vector2f(lx, ly), i); // TODO - CAN BE DONE ONCE!
							Vector3f lightPixelNormal = new Vector3f(lightPlanes.get(i).Normal(lx, ly));
							//lightPixelNormal.x = 2.0f*(lightPixelNormal.x-0.5f); // -1 to 1
							//lightPixelNormal.y = 2.0f*(lightPixelNormal.y-0.5f);
							//lightPixelNormal.z = 2.0f*(lightPixelNormal.z-0.5f);
							//lightPixelNormal.normalise();
							
							//WRONG: lightDir = pixel to pixel direction!
							
							// pixel normal => surface normal
							// compared to direction of surface to light surface..
							// doesn't take into account the orientation of the light surface...
							
							Vector3f lightDir = new Vector3f();
							lightDir.x = lightPixelPos.x - pixelPos.x; //swapped; pixelPos ... position of the current pixel
							lightDir.y = lightPixelPos.y - pixelPos.y;
							lightDir.z = lightPixelPos.z - pixelPos.z;
							lightDir.normalise();
							
							// from g-buffer frag
							//float d = length(pass_LightSpacePosition.xyz)/45;
							//d = 1.0/pow(d,2.0);
							
							// only for the image pixel
							float ndotl = Vector3f.dot(lightDir, pixelNormal); //lightPixelNormal
							if (ndotl < 0) {
								ndotl = 0;
							}
							if (ndotl > 1) {
								ndotl = 1;
							}
							
							// for the light pixel
							float ndotl_light = Vector3f.dot((Vector3f)lightDir.negate(), lightPixelNormal);
							if (ndotl_light < 0) {
								ndotl_light = 0;
							}
							if (ndotl_light > 1) {
								ndotl_light = 1;
							}
							
							Vector3f pPos = new Vector3f(pixelPos.x, pixelPos.y, pixelPos.z);
							Vector3f lPPos = new Vector3f(lightPixelPos.x, lightPixelPos.y, lightPixelPos.z);
							Vector3f dst = new Vector3f();
							Vector3f.sub(pPos, lPPos, dst);
							double len = (double)dst.length()*1.0;
							double attenuation = 1.0/(1.0f + len*len);
							
							// surface color times light
							newLight[0] += (indirect_factor * attenuation * ndotl * ndotl_light * lightPixelColor.x * pixelColor.x);
							newLight[1] += (indirect_factor * attenuation * ndotl * ndotl_light * lightPixelColor.y * pixelColor.y);
							newLight[2] += (indirect_factor * attenuation * ndotl * ndotl_light * lightPixelColor.z * pixelColor.z);
							
							//newLight[0] += ndotl*d/5.0f;
							//newLight[1] += ndotl*d/5.0f;
							//newLight[2] += ndotl*d/5.0f;
							
							//newLight[0] += d/15.0f; // LOOKS COOL (distance sum) - maybe artifact at shadow/object boundaries?
							//newLight[1] += d/15.0f;
							//newLight[2] += d/15.0f;
							
						}
					}
					
					double size = c_size[0] * c_size[1]; // for normalization
					
					size /= 100.0; //100, 300, 500
					
					newLight[0] /= size;
					newLight[1] /= size;
					newLight[2] /= size;
					
					lightSum[0] += newLight[0];
					lightSum[1] += newLight[1];
					lightSum[2] += newLight[2];
					
				}
				
				double[] light = new double[]{
						directIllum.get(0)[x][y].x + lightSum[0], 
						directIllum.get(0)[x][y].y + lightSum[1], 
						directIllum.get(0)[x][y].z + lightSum[2]};
				
				//light = new double[]{		//DEBUG DISPLAY for SINGLOUTPUT
				//		1.0f*lightSum[0], 
				//		1.0f*lightSum[1], 
				//		1.0f*lightSum[2]};
				
				//if (light[0] > 0.1) {
				//	result[x][y] = new Vector4f(1.0f, 0.0f, 0.0f, 1.0f);
				//} else {
					result[x][y] = new Vector4f((float)light[0], (float)light[1], (float)light[2], 1.0f);
				//}
				
				//directIllum.get(0)[x][y].x, directIllum.get(0)[x][y].y, directIllum.get(0)[x][y].z
				
			}
			
			System.out.println("row " + y);
			
		}
		
		return result;
	}
	
	
	/**
	 * 
	 * This is the version using a quad tree for light intensities.
	 * 
	 * @return
	 */
	private Vector4f[][] optimizedCalcIndirectIllumSingleQTInt() {
		
		Vector4f[][] result = new Vector4f[c_size[0]][c_size[1]];
		
		ArrayList<QuadTree<Float>> qts = new ArrayList<QuadTree<Float>>();
		
		// pre-calculate position arrays from depths
		for (int i = 0; i < directIllumSingle.size(); i++) {
			qts.add(new QuadTree<Float>(lightPlanes.get(i).getLightFloat()));
		}
		
		//System.out.println("Positions calculated");
		
		// for all pixels
		for (int y = 0; y < c_size[1]; y++) {
			for (int x = 0; x < c_size[0]; x++) {
				
				double[] lightSum = new double[]{0, 0, 0};
				
				//Vector4f newLight = new Vector4f(directIllum.get(0)[x][y]); // direct illumination
				
				Vector4f pixelColor = new Vector4f(cameraPlane.Albedo(x, y));
				Vector4f pixelPos = getWorldSpacePosition(new Vector2f(x, y));
				Vector3f pixelNormal = new Vector3f(cameraPlane.Normal(x, y));
				
				// indirect
				// over CAMERA
				
				// over all LIGHT_SOURCES directIllumSingle.size()
				for (int i = 0; i < directIllumSingle.size(); i++) {
					
					double[] newLight = new double[]{0, 0, 0};
					
					// 1000
					newLight = qts.get(i).generateSummationOverImage(1000, directIllumSingle.get(i), lightPlanes.get(i), lightPlanePositions.get(i), pixelColor, pixelPos, pixelNormal);
					
					lightSum[0] += newLight[0];
					lightSum[1] += newLight[1];
					lightSum[2] += newLight[2];
					
				}
				
				double[] light = new double[]{
						directIllum.get(0)[x][y].x + lightSum[0], 
						directIllum.get(0)[x][y].y + lightSum[1], 
						directIllum.get(0)[x][y].z + lightSum[2]};
				
				/*light = new double[]{		//DEBUG DISPLAY for SINGLOUTPUT
						1.0f*lightSum[0], 
						1.0f*lightSum[1], 
						1.0f*lightSum[2]};*/
				
				//if (light[0] > 0.1) {
				//	result[x][y] = new Vector4f(1.0f, 0.0f, 0.0f, 1.0f);
				//} else {
					result[x][y] = new Vector4f((float)light[0], (float)light[1], (float)light[2], 1.0f);
				//}
				
				//directIllum.get(0)[x][y].x, directIllum.get(0)[x][y].y, directIllum.get(0)[x][y].z
				
			}
			
			System.out.println("row " + y);
			
		}
		
		return result;
	}
	
	/**
	 * This is the version using a quad tree for positions.
	 * 
	 * @param radius_multiplier defines the discard relative to the values
	 * @return
	 */
	private Vector4f[][] optimizedCalcIndirectIllumSingleQTPos(float radius_multiplier) {
		
		Vector4f[][] result = new Vector4f[c_size[0]][c_size[1]];
		
		ArrayList<DistanceQuadTree> dqts = new ArrayList<DistanceQuadTree>();
		
		//new DistanceQuadTree(lightPlanePositions.get(0).getWs_positionArray(), directIllumSingle.get(0))
		
		// pre-calculate position arrays from depths
		for (int i = 0; i < directIllumSingle.size(); i++) {
			dqts.add(new DistanceQuadTree(lightPlanePositions.get(i).getWs_positionArray(), directIllumSingle.get(i)));
		}
		
		//System.out.println("Positions calculated");
		
		// for all pixels
		for (int y = 0; y < c_size[1]; y++) {
			for (int x = 0; x < c_size[0]; x++) {
				
				double[] lightSum = new double[]{0, 0, 0};
				
				//Vector4f newLight = new Vector4f(directIllum.get(0)[x][y]); // direct illumination
				
				Vector4f pixelColor = new Vector4f(cameraPlane.Albedo(x, y));
				Vector4f pixelPos = getWorldSpacePosition(new Vector2f(x, y));
				Vector3f pixelNormal = new Vector3f(cameraPlane.Normal(x, y));
				
				// indirect
				// over CAMERA
				
				// over all LIGHT_SOURCES directIllumSingle.size()
				for (int i = 0; i < directIllumSingle.size(); i++) {
					
					double[] newLight = new double[]{0, 0, 0};
					
					newLight = dqts.get(i).generateSummationOverImage(1000, radius_multiplier, directIllumSingle.get(i), lightPlanes.get(i), lightPlanePositions.get(i), pixelColor, pixelPos, pixelNormal);
					
					/*if (newLight[0] != 0) {
						System.out.println(newLight[0] + "/" + newLight[1] + "/" + newLight[2]);
					}*/
					
					lightSum[0] += newLight[0];
					lightSum[1] += newLight[1];
					lightSum[2] += newLight[2];
					
				}
				
				double[] light = new double[]{
						directIllum.get(0)[x][y].x + lightSum[0], 
						directIllum.get(0)[x][y].y + lightSum[1], 
						directIllum.get(0)[x][y].z + lightSum[2]};
				
				result[x][y] = new Vector4f((float)light[0], (float)light[1], (float)light[2], 1.0f);
				
				/*if (light[0] > 1.1) { // TODO SOMETIMES > 1 ??
					System.out.println(light[0]);
					System.out.println(directIllum.get(0)[x][y].x);
					
					result[x][y] = new Vector4f(0.0f, 0.0f, 0.0f, 1.0f);
				}*/
			}
			
			System.out.println("row " + y);
			
		}
		
		return result;
	}
	
	public float[] calcIllumination() {
		
		startTime = System.nanoTime();
		
		float[] resultArr = new float[4*c_size[0]*c_size[1]];
		
		//result = calcDirect();
		directIllum.add(calcDirect());
		for (int i = 0; i < lightPlanes.size(); i++) {
			//directIllum.add(calcDirectLightSource(i));
			directIllumSingle.add(calcDirectLightSourceSingle(i));
		}
		
		//result = directIllum.get(0);
		
		/**
		 * CALCULATE 2nd bounce
		 */
		
		// pre-calculate position arrays from depths
		for (int i = 0; i < directIllumSingle.size(); i++) {
			lightPlanePositions.add(new ImagePlanePosition<Vector4f, Vector3f, Float, Float>(lightPlanes.get(i)));
			lightPlanePositions.get(i).calculateWorldPositions();
		}
		//result = DataTypeConverter.toVec4(lightPlanePositions.get(0).getWs_positionArray());
		//result = directIllumSingle.get(4);
		
		// calcIndirectIllumSingle // calcIndirectIllumSummed //directIllum.get(1).clone();
		
		
		/*
		32x32
		calcIndirectIllumSingle: 						Runtime 1,313815 Seconds
		optimizedCalcIndirectIllumSingleNoQuadtree: 	Runtime 0,706874 Seconds
		optimizedCalcIndirectIllumSingleQTInt: 			Runtime 0,312525 Seconds
		optimizedCalcIndirectIllumSingleQTPos(32): 		Runtime 0,131249 Seconds
		
		64x64
		calcIndirectIllumSingle: 						Runtime 16,366410 Seconds
		optimizedCalcIndirectIllumSingleNoQuadtree: 	Runtime 9,813519 Seconds
		optimizedCalcIndirectIllumSingleQTInt: 			Runtime 3,880756 Seconds
		optimizedCalcIndirectIllumSingleQTPos(32): 		Runtime 1,340149 Seconds
		
		128x128
		calcIndirectIllumSingle: 						Runtime 282,915735 Seconds
		optimizedCalcIndirectIllumSingleNoQuadtree: 	Runtime 169,288592 Seconds
		optimizedCalcIndirectIllumSingleQTInt: 			Runtime 60,629756 Seconds
		optimizedCalcIndirectIllumSingleQTPos(32): 		Runtime 14,274376 Seconds
		*/
		
		//result = calcIndirectIllumSingle(); 						// 	64x64 ~ 17s											//TODO FIX
		//result = optimizedCalcIndirectIllumSingleNoQuadtree();	// 	64x64 ~ 10s
		//result = optimizedCalcIndirectIllumSingleQTInt();			// 	64x64 ~ 3,8s 			//ACTUAL!!!!!!!!!!!!		//TODO FIX
		// 																64x64 ~ 3,844258 Seconds for 255 radius multiplier
		// 																64x64 ~ 2,578632 Seconds for 128 radius multiplier
		// 																64x64 ~ 1,863936 Seconds for 64 radius multiplier
		// 256x256 ~ Runtime 179,068951 Seconds 
		result = optimizedCalcIndirectIllumSingleQTPos(128.0f); //32.0f
		
		
		//test
		
		//QuadTree<Float> qt = new QuadTree<Float>(lightPlanes.get(0).getLightFloat());
		//Float[][] tmpResult = (Float[][])qt.generateImage(5);
		//result = DataTypeConverter.toVector(tmpResult);
		
		//DistanceQuadTree dqt = new DistanceQuadTree(lightPlanePositions.get(0).getWs_positionArray(), directIllumSingle.get(0));
		//result = dqt.generateRadiusImage(4); 	// DISPLAYS a debug view of the radius of the bounding-sphere quadtree
		//result = dqt.generateMeanImage(4); 		// DISPLAYS a debug view of the mean of the bounding-sphere quadtree
		//result = dqt.getInputLightRadiusImage();
		
		//result = directIllumSingle.get(0);
		
		//result = lightPlanes.get(0).getDepth();
		
		/**
		 * CALCULATE 2nd bounce END
		 */
		
		//result = directIllumSingle.get(4);
		
		//result = calcDirect();
		
		//result = lightPlanes.get(0).getLight();
		
		//System.out.println("done");
		
		// convert to linear
		int index = 0;
		for (int y = 0; y < c_size[1]; y++) {
			for (int x = 0; x < c_size[0]; x++) {
				
				//camera
				resultArr[4*index]   = result[x][y].x;
				resultArr[4*index+1] = result[x][y].y;
				resultArr[4*index+2] = result[x][y].z;
				resultArr[4*index+3] = result[x][y].w;
				
				//light 1
				/*resultArr[4*index]   = lightPlanes.get(0).Albedo(x, y).x;
				resultArr[4*index+1] = lightPlanes.get(0).Albedo(x, y).y;
				resultArr[4*index+2] = lightPlanes.get(0).Albedo(x, y).z;
				resultArr[4*index+3] = lightPlanes.get(0).Albedo(x, y).w;*/
				index++;
				
			}
		}
		
		long duration = System.nanoTime() - startTime;
		final double seconds = ((double)duration / 1000000000);
		
		System.out.format("Runtime %f Seconds", seconds);
		System.out.println();
		
		//resultArr = linearize(resultArr);
		
		return resultArr;
		
	}
	
}
