package modules.render.cpu;

import org.lwjgl.util.vector.Matrix4f;
import org.lwjgl.util.vector.Vector3f;
import org.lwjgl.util.vector.Vector4f;

/**
 * Class for an ImagePlane:
 * This is an array of pixels with camera parameters attached.
 * 
 * Albedo, Normal, Distance, Light
 * 
 * @author DarkMatt
 *
 * @param <T>
 */
public class ImagePlane<A, N, D, L> {

	private int sx, sy;
	
	private A[][] albedo;
	private N[][] normal;
	private D[][] depth;
	private L[][] light;
	private D[][] depth2;
	
	private Matrix4f viewMatrix;
	private Matrix4f invViewMatrix;
	private Matrix4f projMatrix;
	private Matrix4f invProjMatrix;
	private Matrix4f viewProjMatrix;
	
	@SuppressWarnings("unchecked")
	public ImagePlane(int sx, int sy) {
		this.sx = sx;
		this.sy = sy;
		albedo = 	(A[][]) new Object[sx][sy];
		normal = 	(N[][]) new Object[sx][sy];
		depth = 	(D[][]) new Object[sx][sy];
		light = 	(L[][]) new Object[sx][sy];
		depth2 = 	(D[][]) new Object[2*sx][2*sy];
		
		viewMatrix = new Matrix4f();
		projMatrix = new Matrix4f();
	}
	
	public void setAlbedo(int x, int y, A value) {
		albedo[x][y] = value;
	}
	
	public void setNormal(int x, int y, N value) {
		normal[x][y] = value;
	}
	
	public void setDepth(int x, int y, D value) {
		depth[x][y] = value;
	}
	
	public void setLight(int x, int y, L value) {
		light[x][y] = value;
	}
	
	public void setDepth2(int x, int y, D value) {
		depth2[x][y] = value;
	}
	
	public A Albedo(int x, int y) {
		return albedo[x][y];
	}
	
	public N Normal(int x, int y) {
		return normal[x][y];
	}
	
	public Vector4f[][] getAlbedo() {
		
		Vector4f[][] out = new Vector4f[sx][sy];
		
		for (int y = 0; y < sy; y++) {
			for (int x = 0; x < sx; x++) {
				
				Vector4f a = (Vector4f)(albedo[x][y]);
				
				out[x][y] = new Vector4f(a.x, a.y, a.z, a.w);
				
			}
		}
		
		return out;
	}
	
	public Vector4f[][] getNormals() {
		
		Vector4f[][] out = new Vector4f[sx][sy];
		
		for (int y = 0; y < sy; y++) {
			for (int x = 0; x < sx; x++) {
				
				Vector3f n = (Vector3f)(normal[x][y]);
				
				out[x][y] = new Vector4f(n.x, n.y, n.z, 1.0f);
				
			}
		}
		
		return out;
	}
	
	public Vector4f[][] getDepth() {
		
		Vector4f[][] out = new Vector4f[sx][sy];
		
		for (int y = 0; y < sy; y++) {
			for (int x = 0; x < sx; x++) {
				
				float d = (Float)(depth[x][y]);
				
				out[x][y] = new Vector4f(d, d, d, 1.0f);
				
			}
		}
		
		return out;
	}
	
	public Vector4f[][] getLight() {
		
		Vector4f[][] out = new Vector4f[sx][sy];
		
		for (int y = 0; y < sy; y++) {
			for (int x = 0; x < sx; x++) {
				
				float l = (Float)(light[x][y]);
				
				out[x][y] = new Vector4f(l, l, l, 1.0f);
				
			}
		}
		
		return out;
	}
	
	// override for just getting float values
	public Float[][] getLightFloat() {
		
		Float[][] out = new Float[sx][sy];
		
		for (int y = 0; y < sy; y++) {
			for (int x = 0; x < sx; x++) {
				
				float l = (Float)(light[x][y]);
				
				out[x][y] = l;
				
			}
		}
		
		return out;
	}
	
	public D Depth(int x, int y) {
		return depth[x][y];
	}
	
	public L Light(int x, int y) {
		return light[x][y];
	}
	
	public D Depth2(int x, int y) {
		return depth2[x][y];
	}
	
	/**
	 * Sets view matrix. Automatically calculates the inverse view matrix.
	 * 
	 * @param viewMatrix
	 */
	public void setViewMatrix(Matrix4f viewMatrix) {
		this.viewMatrix = viewMatrix;
		// inverse of the matrix
		this.invViewMatrix = (Matrix4f)new Matrix4f(viewMatrix).invert();
		
		this.calculateViewProjMatrix(); // can be slightly redundant
	}
	
	/**
	 * Sets inverse view matrix. Automatically calculates the view matrix.
	 * 
	 * @param viewMatrix
	 */
	public void setInvViewMatrix(Matrix4f invViewMatrix) {
		this.invViewMatrix = invViewMatrix;
		// inverse of the matrix
		this.viewMatrix = (Matrix4f)new Matrix4f(invViewMatrix).invert();
		
		this.calculateViewProjMatrix(); // can be slightly redundant
	}
	
	/**
	 * Returns the view matrix.
	 * 
	 * @return
	 */
	public Matrix4f getViewMatrix() {
		return viewMatrix;
	}
	
	/**
	 * Returns the inverse view matrix.
	 * 
	 * @return
	 */
	public Matrix4f getInvViewMatrix() {
		return invViewMatrix;
	}
	
	/**
	 * Sets projection matrix. Automatically calculates the inverse projection matrix.
	 * 
	 * @param projMatrix
	 */
	public void setProjMatrix(Matrix4f projMatrix) {
		this.projMatrix = projMatrix;
		// inverse of the matrix
		this.invProjMatrix = (Matrix4f)new Matrix4f(projMatrix).invert();
		
		this.calculateViewProjMatrix();
	}
	
	/**
	 * Sets inverse projection matrix. Automatically calculates the projection matrix.
	 * 
	 * @param projMatrix
	 */
	public void setInvProjMatrix(Matrix4f invProjMatrix) {
		this.invProjMatrix = invProjMatrix;
		// inverse of the matrix
		this.projMatrix = (Matrix4f)new Matrix4f(invProjMatrix).invert();
		
		this.calculateViewProjMatrix();
	}
	
	/**
	 * Returns the projection matrix.
	 * 
	 * @return
	 */
	public Matrix4f getProjMatrix() {
		return projMatrix;
	}
	
	/**
	 * Returns the inverse projection matrix.
	 * 
	 * @return
	 */
	public Matrix4f getInvProjMatrix() {
		return invProjMatrix;
	}
	
	public Matrix4f getViewProjMatrix() {
		return viewProjMatrix;
	}
	
	private void calculateViewProjMatrix() {
		Matrix4f toAdd = new Matrix4f();
		Matrix4f.mul(projMatrix, viewMatrix, toAdd);
		
		viewProjMatrix = toAdd;
	}
	
	public int getSx() {
		return sx;
	}
	
	public int getSy() {
		return sy;
	}
	
}
