package modules.render.cpu;

import org.lwjgl.util.vector.Vector3f;
import org.lwjgl.util.vector.Vector4f;

public class QuadTree <T> {

	private float indirect_factor = 0.75f;
	private double normalizationFactor = 1.0/100.0; // 1.0/100.0
	
	private boolean debug = true; // shows cell size of quad tree
	
	private int sx;
	private int sy;
	/** Needed for initialization */
	private T[][] inputData;
	
	//private float counter = 0;
	
	private class Node<G> {
		public G value;
		public Node<G> lt;
		public Node<G> rt;
		public Node<G> lb;
		public Node<G> rb;
	}
	private Node<T> root;
	
	private int localMaxLevel;
	
	public QuadTree(T[][] inputData) {
		this.inputData = inputData;
		
		sx = inputData.length;
		sy = inputData[0].length;
		
		normalizationFactor *= sx * sy;
		
		if (!isSquarePowerOfTwo(inputData)) System.out.println("Error: non power of two quad-tree");
		
		int maxlevel = (int)(Math.log(inputData.length)/Math.log(2));
		localMaxLevel = maxlevel;
		
		root = traverse(maxlevel, 1, 0, 0);
	
		//for (int i = 0; i < maxlevel; i++) {
			recursivePrune(root);
		//}
		
		
		//System.out.println(root);
	}
	
	private Node<T> traverse(int level, int size, int px, int py) {
		
		Node<T> n = new Node<T>(); // lt, rt, lb, rb are null if no children
		
		// children
		int tLevel = level-1;
		if (tLevel >= 0) {
			n.lt = traverse(tLevel, 2*size, 2*px  , 2*py  );
			n.rt = traverse(tLevel, 2*size, 2*px+1, 2*py  );
			n.lb = traverse(tLevel, 2*size, 2*px  , 2*py+1);
			n.rb = traverse(tLevel, 2*size, 2*px+1, 2*py+1);
			
			perNodeCalculation(n);
		} else {
			n.value = inputData[px][py];
		}
		
		return n;
	}
	
	/** Stores maximum of the child nodes in the current */
	private void perNodeCalculation(Node<T> in) {
		
		// stores the maximum of the child nodes in the current node  //TODO performance?
		if (Float.class.isAssignableFrom(in.lt.value.getClass())) {
		
			T maximum = in.lt.value;
			
			if ((Float)in.rt.value > (Float)maximum) {
				maximum = in.rt.value;
			}
			if ((Float)in.lb.value > (Float)maximum) {
				maximum = in.lb.value;
			}
			if ((Float)in.rb.value > (Float)maximum) {
				maximum = in.rb.value;
			}
			
			in.value = maximum;
			
		} else {
			System.out.println("Error - non Float");
		}
		
	}
	
	private void recursivePrune(Node<T> node) {
		
		if(node.lb != null) recursivePrune(node.lb);
		if(node.lt != null) recursivePrune(node.lt);
		if(node.rb != null) recursivePrune(node.rb);
		if(node.rt != null) recursivePrune(node.rt);
		
		if (node.lb != null && node.lt != null && node.rb != null && node.rt != null) {
			if(	(Float)node.lb.value == 0.0f && (Float)node.lt.value == 0.0f && 
				(Float)node.rb.value == 0.0f && (Float)node.rt.value == 0.0f) {
				
				node.lb = null;
				node.lt = null;
				node.rb = null;
				node.rt = null;
			}
		}
		
	}
	
	/** Generates an image from the quad-tree with a specified limit to the recursion */
	public T[][] generateImage(int maximumLevel) {
		
		T[][] out = inputData.clone();
		
		recursiveGeneration(root, out, maximumLevel, 0, 0, sx, sy);
		
		return out;
	}
	
	/** Recursive method to generate a 2d array representing the quad-tree (the value debug allows the cell borers to be displayed) */
	public void recursiveGeneration(Node<T> node, T[][] data, int level, int minX, int minY, int maxX, int maxY) {
		
		if (level > 0 && node != null) {
			
			int centerX = minX + (maxX-minX)/2;
			int centerY = minY + (maxY-minY)/2;
			
			if (node.lt != null && node.lb != null && node.rt != null && node.rb != null) {
				recursiveGeneration(node.lt, data, level-1, minX, minY, centerX, centerY);
				recursiveGeneration(node.lb, data, level-1, minX, centerY, centerX, maxY);
				recursiveGeneration(node.rt, data, level-1, centerX, minY, maxX, centerY);
				recursiveGeneration(node.rb, data, level-1, centerX, centerY, maxX, maxY);
			} else if (debug) {
				for (int x = minX; x < maxX; x++) {
					data[x][maxY-1] = (T)(Float)0.5f;
				}
				for (int y = minY; y < maxY; y++) {
					data[maxX-1][y] = (T)(Float)0.5f;
				}
			}
			
		} else {
			
			//System.out.println(minX + "/" + maxX + ":" + minY + "/" + maxY);
			//counter += 0.05f;
			
			if (debug) {
				
				for (int y = minY; y < maxY-1; y++) {
					for (int x = minX; x < maxX-1; x++) {
						//data[x][y] = (T)(Float)counter;//node.value;
						data[x][y] = node.value;
					}
				}
				
				for (int x = minX; x < maxX; x++) {
					data[x][maxY-1] = (T)(Float)0.5f;//node.value;
				}
				for (int y = minY; y < maxY; y++) {
					data[maxX-1][y] = (T)(Float)0.5f;//node.value;
				}
				
			} else {
				
				for (int y = minY; y < maxY; y++) {
					for (int x = minX; x < maxX; x++) {
						//data[x][y] = (T)(Float)counter;//node.value;
						data[x][y] = node.value;
					}
				}
				
			}
			
			
		}
		
	}
	
	/** Generates the summation over all pixels recursively */
	public double[] generateSummationOverImage(int maximumLevel,
			Vector4f[][] directIllumSingle, 
			ImagePlane<Vector4f, Vector3f, Float, Float> lightPlanes,
			ImagePlanePosition<Vector4f, Vector3f, Float, Float> lightPlanePositions,
			Vector4f pixelColor, Vector4f pixelPos, Vector3f pixelNormal) {
		
		double[] out = new double[]{0.0, 0.0, 0.0};
		
		if (maximumLevel > localMaxLevel) {
			maximumLevel = localMaxLevel;
		}
		
		recursiveSum(out, root, maximumLevel, 0, 0, sx, sy, directIllumSingle, lightPlanes, lightPlanePositions, pixelColor, pixelPos, pixelNormal);
		
		out[0] = (float)(out[0]/normalizationFactor);
		out[1] = (float)(out[1]/normalizationFactor);
		out[2] = (float)(out[2]/normalizationFactor);
		
		return out;
	}
	
	public void recursiveSum(double[] value, Node<T> node, int level, int minX, int minY, int maxX, int maxY,
			Vector4f[][] directIllumSingle, 
			ImagePlane<Vector4f, Vector3f, Float, Float> lightPlanes,
			ImagePlanePosition<Vector4f, Vector3f, Float, Float> lightPlanePositions,
			Vector4f pixelColor, Vector4f pixelPos, Vector3f pixelNormal) {
		
		if (level > 0 && node != null) {
			
			int centerX = minX + (maxX-minX)/2;
			int centerY = minY + (maxY-minY)/2;
			
			if (node.lt != null && node.lb != null && node.rt != null && node.rb != null) {
				recursiveSum(value, node.lt, level-1, minX, minY, centerX, centerY, directIllumSingle, lightPlanes, lightPlanePositions, pixelColor, pixelPos, pixelNormal);
				recursiveSum(value, node.lb, level-1, minX, centerY, centerX, maxY, directIllumSingle, lightPlanes, lightPlanePositions, pixelColor, pixelPos, pixelNormal);
				recursiveSum(value, node.rt, level-1, centerX, minY, maxX, centerY, directIllumSingle, lightPlanes, lightPlanePositions, pixelColor, pixelPos, pixelNormal);
				recursiveSum(value, node.rb, level-1, centerX, centerY, maxX, maxY, directIllumSingle, lightPlanes, lightPlanePositions, pixelColor, pixelPos, pixelNormal);
			}
			
		} else {
			
			// summation over all pixels inside a cell (not executed for zero-valued cells)
			/*for (int y = minY; y < maxY; y++) {
				for (int x = minX; x < maxX; x++) {
					//data[x][y] = (T)(Float)counter;//node.value;
					//data[x][y] = node.value;
					
					value[0] += 1.0/normalizationFactor;
					value[1] += 1.0/normalizationFactor;
					value[2] += 1.0/normalizationFactor;
					
				}
			}*/
			
			//System.out.println("leaf");
			
			recursiveSumLeaf(value, minX, minY, maxX, maxY, 
					directIllumSingle, lightPlanes, lightPlanePositions, pixelColor, pixelPos, pixelNormal);
			
		}
		
	}
	
	/**
	 * 
	 * @param newLight ... the value that is incremented (summed over)
	 * @param minX
	 * @param minY
	 * @param maxX
	 * @param maxY
	 * @param directIllumSingle (at i)
	 * @param lightPlanePositions (at i)
	 */
	private void recursiveSumLeaf(double[] newLight, int minX, int minY, int maxX, int maxY, 
			Vector4f[][] directIllumSingle, 
			ImagePlane<Vector4f, Vector3f, Float, Float> lightPlanes,
			ImagePlanePosition<Vector4f, Vector3f, Float, Float> lightPlanePositions,
			Vector4f pixelColor, Vector4f pixelPos, Vector3f pixelNormal) {
		
		// over all pixels
		for (int ly = minY; ly < maxY; ly++) {
			for (int lx = minX; lx < maxX; lx++) {
				
				Vector4f lightPixelColor = new Vector4f(directIllumSingle[lx][ly]);
				Vector3f lightPixelPos = lightPlanePositions.getWs_position(lx, ly);
				Vector3f lightPixelNormal = new Vector3f(lightPlanes.Normal(lx, ly));
				
				Vector3f lightDir = new Vector3f();
				lightDir.x = lightPixelPos.x - pixelPos.x;
				lightDir.y = lightPixelPos.y - pixelPos.y;
				lightDir.z = lightPixelPos.z - pixelPos.z;
				lightDir.normalise();
				
				float ndotl = Vector3f.dot(lightDir, pixelNormal);
				if (ndotl < 0) {
					ndotl = 0;
				}
				if (ndotl > 1) {
					ndotl = 1;
				}
				
				// for the light pixel
				float ndotl_light = Vector3f.dot((Vector3f)lightDir.negate(), lightPixelNormal);
				if (ndotl_light < 0) {
					ndotl_light = 0;
				}
				if (ndotl_light > 1) {
					ndotl_light = 1;
				}
				
				Vector3f pPos = new Vector3f(pixelPos.x, pixelPos.y, pixelPos.z);
				Vector3f lPPos = new Vector3f(lightPixelPos.x, lightPixelPos.y, lightPixelPos.z);
				Vector3f dst = new Vector3f();
				Vector3f.sub(pPos, lPPos, dst);
				double len = (double)dst.length();
				double attenuation = 1.0/(1.0f+Math.pow(len, 2.0)); // pow 1.5
				
				// surface color times light
				newLight[0] += (indirect_factor * attenuation * ndotl * ndotl_light * lightPixelColor.x * pixelColor.x);
				newLight[1] += (indirect_factor * attenuation * ndotl * ndotl_light * lightPixelColor.y * pixelColor.y);
				newLight[2] += (indirect_factor * attenuation * ndotl * ndotl_light * lightPixelColor.z * pixelColor.z);
				
			}
		}
		
	}
	
	/** Only accepts square */
	private boolean isSquarePowerOfTwo(T[][] array){
		int size = array.length;
		if (!isPowerOfTwo(size)) return false;
		for (int i = 0; i < array.length; i++) if (array[i].length != size) return false;
		return true;
	}
	
	/** http://javarevisited.blogspot.co.at/2013/05/how-to-check-if-integer-number-is-power-of-two-example.html */
    private static boolean isPowerOfTwo(int value){
        return ((value & (value -1)) == 0);
    }
	
}
