package modules.render.cpu.boundingsphere;

import java.util.ArrayList;

import modules.render.cpu.ImagePlane;
import modules.render.cpu.ImagePlanePosition;

import org.lwjgl.util.vector.Vector3f;
import org.lwjgl.util.vector.Vector4f;

/**
 * This implementation of a quad tree that stores bounding spheres at its nodes.
 * 
 * @author DarkMatt
 *
 */
public class DistanceQuadTree {

	//private static final float RADIUSMULTIPLIER = 32.0f; //64.0f, 255.0f -- also depends on the normalization factor :(
	
	/**
	 * for 64x64
	 * RADIUSMULTIPLIER:
	 * 16 	- visibly wrong - Runtime 0,775392 Seconds
	 * 32+ looks OK
	 * 32 	- Runtime 1,435191 Seconds - normalizationFactor = 1.0/200.0
	 * 64 	- Runtime 1,868093 Seconds - normalizationFactor = 1.0/150.0
	 * 128 	- Runtime 2,533527 Seconds - normalizationFactor = 1.0/100.0
	 * 255	- Runtime 3,585770 Seconds
	 * 512	- Runtime 3,729406 Seconds
	 * 1024	- Runtime 3,831920 Seconds
	 * 2048 - Runtime 4,229951 Seconds
	 * 
	 * 5000 - Runtime 4,666921 Seconds
	 * 
	 * 10000		- Runtime 4,554764 Seconds
	 * 100000		- Runtime 4,792590 Seconds
	 * 100000000	- Runtime 4,789316 Seconds
	 * 1000000000 	- Runtime 4,660827 Seconds
	 */
	
	private static final float minValue = 0.001f;
	// 1/100 -> 1/150 .. scene 1, 2; 1/200 .. scene 0
	
	// 750 
	private double normalizationFactor = 1.0/200.0; //used: 1.0/100.0 //1.0/200.0 //1.0/100.0 - smaller denominator means darker
	
	float indirect_factor = 1.0f; //0.75f
	
	private int localMaxLevel;
	
	private int sx;
	private int sy;
	private BoundingSphere[][] data;
	private Vector3f[][] positionData;
	private Vector4f[][] directIllumination;
	
	private class Node {
		public BoundingSphere value;
		public Node lt;
		public Node rt;
		public Node lb;
		public Node rb;
	}
	private Node root;
	
	/**
	 * Initalizes the quad tree with the data necessary for generation.
	 * 
	 * @param positionData
	 */
	public DistanceQuadTree(Vector3f[][] positionData, Vector4f[][] directIllumination) {
		this.positionData = positionData;
		this.directIllumination = directIllumination;
		
		sx = positionData.length;
		sy = positionData[0].length;
		
		normalizationFactor *= sx * sy;
		
		if (!isSquarePowerOfTwo(positionData)) System.out.println("Error: non power of two quad-tree");
		
		data = generateData();
		
		int maxlevel = (int)(Math.log(positionData.length)/Math.log(2));
		localMaxLevel = maxlevel;
		
		// traverse and fill tree
		root = buildTree(maxlevel, 1, 0, 0);
		
		// generate bounding-spheres
		generateBoundingSpheres(root);
		
		// too small sized bounding-spheres need to be removed from the quad tree -> not necessary
		
		
	}
	
	private BoundingSphere[][] generateData() {
		
		BoundingSphere[][] out = new BoundingSphere[sx][sy];
		
		for (int y = 0; y < sy; y++) {
			for (int x = 0; x < sx; x++) {
				
				Vector3f pos = positionData[x][y];
				
				// find maximal color intensity
				float maxIntensity = 0.0f;
				if (directIllumination[x][y].x > maxIntensity) {
					maxIntensity = directIllumination[x][y].x;
				}
				if (directIllumination[x][y].y > maxIntensity) {
					maxIntensity = directIllumination[x][y].y;
				}
				if (directIllumination[x][y].z > maxIntensity) {
					maxIntensity = directIllumination[x][y].z;
				}
				
				float radius = 0.0f;
				
				float atten = 0.0f;
				if (maxIntensity > 0.00001f) {
					atten = (maxIntensity * maxIntensity); //TODO test
				}
				
				if (atten > minValue) {
					radius = atten;
				}
				
				out[x][y] = new BoundingSphere(pos, radius);
				
			}
		}
		
		return out;
	}
	
	/**
	 * Recursively traverse tree + for every node calculate bounding-sphere (except for leaves)
	 * 
	 * @param in
	 */
	private void generateBoundingSpheres(Node in) {
		if (in.lt != null && in.rt != null && in.lb != null && in.rb != null) {
			generateBoundingSpheres(in.lt);
			generateBoundingSpheres(in.rt);
			generateBoundingSpheres(in.lb);
			generateBoundingSpheres(in.rb);
			
			// generates the bounding sphere for every non leaf node
			in.value = generateBoundingSphere(in);
		}
	}
	
	private BoundingSphere generateBoundingSphere(Node in) {
		
		ArrayList<BoundingSphere> bspheres = new ArrayList<BoundingSphere>();
		generateBoundingSphereChildrenList(in, bspheres);
		
		AABBBoundingSphere aabbbs = new AABBBoundingSphere();
		aabbbs.addBoundingSpheres(bspheres);
		aabbbs.calculateBoundingSphere();
		
		return aabbbs.getBoundingSphereResult();
	}
	
	/**
	 * Returns a filled list which is passed in the method as parameter.
	 * 
	 * @param in
	 * @param bspheres
	 */
	private void generateBoundingSphereChildrenList(Node in, ArrayList<BoundingSphere> bspheres) {
		
		if (in.lt != null && in.rt != null && in.lb != null && in.rb != null) {
			generateBoundingSphereChildrenList(in.lt, bspheres);
			generateBoundingSphereChildrenList(in.rt, bspheres);
			generateBoundingSphereChildrenList(in.lb, bspheres);
			generateBoundingSphereChildrenList(in.rb, bspheres);
		} else {
			// leaf
			
			if (in.value.getRadius() > minValue) {
				bspheres.add(in.value);
			}
		}
		
	}
	
	/*private BoundingSphere boundingSphereRecursion(Node in) {
		
		float[] mean = new float[]{0.0f, 0.0f, 0.0f};
		
		if (in.lt != null && in.rt != null && in.lb != null && in.rb != null) {
		
			BoundingSphere b1 = boundingSphereRecursion(in.lb);
			BoundingSphere b2 = boundingSphereRecursion(in.rb);
			BoundingSphere b3 = boundingSphereRecursion(in.lt);
			BoundingSphere b4 = boundingSphereRecursion(in.rt);
			
			mean[0] = (m1.x + m2.x + m3.x + m4.x)/4.0f;
			mean[1] = (m1.y + m2.y + m3.y + m4.y)/4.0f;
			mean[2] = (m1.z + m2.z + m3.z + m4.z)/4.0f;
			
		} else {
			// leaf
			
			mean[0] = in.value.getPos().x;
			mean[1] = in.value.getPos().y;
			mean[2] = in.value.getPos().z;
			
		}
		
		return new Vector3f(mean[0], mean[1], mean[2]);
		
	}*/
	
	/**
	 * 
	 * 
	 * @param in
	 * @return
	 */
	@Deprecated
	private BoundingSphere generateBoundingSphereOld(Node in) {
		
		Vector3f mean = meanRecursion(in);
		
		float radius = radiusRecursion(mean, in);
		
		return new BoundingSphere(mean, radius);
	}
	
	/**
	 * To calculate the mean of all child node positions
	 * 
	 * @param mean
	 * @param in
	 */
	private Vector3f meanRecursion(Node in) {
		
		float[] mean = new float[]{0.0f, 0.0f, 0.0f};
		
		if (in.lt != null && in.rt != null && in.lb != null && in.rb != null) {
		
			Vector3f m1 = meanRecursion(in.lb);
			Vector3f m2 = meanRecursion(in.rb);
			Vector3f m3 = meanRecursion(in.lt);
			Vector3f m4 = meanRecursion(in.rt);
			
			mean[0] = (m1.x + m2.x + m3.x + m4.x)/4.0f;
			mean[1] = (m1.y + m2.y + m3.y + m4.y)/4.0f;
			mean[2] = (m1.z + m2.z + m3.z + m4.z)/4.0f;
			
		} else {
			// leaf
			
			mean[0] = in.value.getPos().x;
			mean[1] = in.value.getPos().y;
			mean[2] = in.value.getPos().z;
			
		}
		
		return new Vector3f(mean[0], mean[1], mean[2]);
		
	}
	
	/**
	 * To calculate the radius of all child node positions
	 * 
	 * @param mean
	 * @param in
	 */
	private float radiusRecursion(Vector3f mean, Node in) {
		
		float radius = 0.0f;
		
		if (in.lt != null && in.rt != null && in.lb != null && in.rb != null) {
		
			float d1 = radiusRecursion(mean, in.lb);
			float d2 = radiusRecursion(mean, in.rb);
			float d3 = radiusRecursion(mean, in.lt);
			float d4 = radiusRecursion(mean, in.rt);
			
			// maximum
			radius = d1;
			if (d2 > radius) {
				radius = d2;
			}
			if (d3 > radius) {
				radius = d3;
			}
			if (d4 > radius) {
				radius = d4;
			}
			
			//mean[0] = (m1[0] + m2[0] + m3[0] + m4[0])/4.0f;
			//mean[1] = (m1[1] + m2[1] + m3[1] + m4[1])/4.0f;
			//mean[2] = (m1[2] + m2[2] + m3[2] + m4[2])/4.0f;
			
		} else {
			// leaf
			
			radius = Vector3f.sub(mean, in.value.getPos(), null).length(); // old
			//radius = in.value.getRadius();
			
		}
		
		return radius;
		
	}
	
	private float distance(float[] p1, float[] p2) {
		return (float)Math.sqrt((p1[0]-p2[0])*(p1[0]-p2[0]) + (p1[1]-p2[1])*(p1[1]-p2[1]) + (p1[2]-p2[2])*(p1[2]-p2[2]));
	}
	
	/**
	 * Builds the tree and fills the leaf nodes
	 * 
	 * @param level
	 * @param size
	 * @param px
	 * @param py
	 * @return
	 */
	private Node buildTree(int level, int size, int px, int py) {
		
		Node n = new Node(); // lt, rt, lb, rb are null if no children
		
		// children
		int tLevel = level-1;
		if (tLevel >= 0) {
			n.lt = buildTree(tLevel, 2*size, 2*px  , 2*py  );
			n.rt = buildTree(tLevel, 2*size, 2*px+1, 2*py  );
			n.lb = buildTree(tLevel, 2*size, 2*px  , 2*py+1);
			n.rb = buildTree(tLevel, 2*size, 2*px+1, 2*py+1);
			
			//perNodeCalculation(n, tLevel);
		} else {
			// add leaf nodes
			n.value = new BoundingSphere(data[px][py]);
		}
		
		return n;
	}
	
	/** Only accepts square */
	private boolean isSquarePowerOfTwo(Vector3f[][] array){
		int size = array.length;
		if (!isPowerOfTwo(size)) return false;
		for (int i = 0; i < array.length; i++) if (array[i].length != size) return false;
		return true;
	}
	
	/** http://javarevisited.blogspot.co.at/2013/05/how-to-check-if-integer-number-is-power-of-two-example.html */
    private static boolean isPowerOfTwo(int value){
        return ((value & (value -1)) == 0);
    }
    
    
    // ---------------------------- DEBUG ------------------------------------
    
    boolean debug = false;
    
    /** Generates an image from the quad-tree with a specified limit to the recursion */
	public Vector4f[][] generateMeanImage(int maximumLevel) {
		
		Vector4f[][] out = new Vector4f[sx][sy];
		
		recursiveGenerationMean(root, out, maximumLevel, 0, 0, sx, sy);
		
		return out;
	}
	
	/** Recursive method to generate a 2d array representing the quad-tree (the value debug allows the cell borers to be displayed) */
	public void recursiveGenerationMean(Node node, Vector4f[][] data, int level, int minX, int minY, int maxX, int maxY) {
		
		if (level > 0 && node != null) {
			
			int centerX = minX + (maxX-minX)/2;
			int centerY = minY + (maxY-minY)/2;
			
			if (node.lt != null && node.lb != null && node.rt != null && node.rb != null) {
				recursiveGenerationMean(node.lt, data, level-1, minX, minY, centerX, centerY);
				recursiveGenerationMean(node.lb, data, level-1, minX, centerY, centerX, maxY);
				recursiveGenerationMean(node.rt, data, level-1, centerX, minY, maxX, centerY);
				recursiveGenerationMean(node.rb, data, level-1, centerX, centerY, maxX, maxY);
			} else if (debug) {
				for (int x = minX; x < maxX; x++) {
					data[x][maxY-1] = new Vector4f(0.5f, 0.5f, 0.5f, 1.0f);
				}
				for (int y = minY; y < maxY; y++) {
					data[maxX-1][y] = new Vector4f(0.5f, 0.5f, 0.5f, 1.0f);
				}
			}
			
		} else {
			
			//System.out.println(minX + "/" + maxX + ":" + minY + "/" + maxY);
			//counter += 0.05f;
			
			if (debug) {
				
				for (int y = minY; y < maxY-1; y++) {
					for (int x = minX; x < maxX-1; x++) {
						//data[x][y] = (T)(Float)counter;//node.value;
						data[x][y] = new Vector4f(node.value.getPos().x, node.value.getPos().y, node.value.getPos().z, 1.0f);
					}
				}
				
				for (int x = minX; x < maxX; x++) {
					data[x][maxY-1] = new Vector4f(0.5f, 0.5f, 0.5f, 1.0f);//node.value;
				}
				for (int y = minY; y < maxY; y++) {
					data[maxX-1][y] = new Vector4f(0.5f, 0.5f, 0.5f, 1.0f);//node.value;
				}
				
			} else {
				
				for (int y = minY; y < maxY; y++) {
					for (int x = minX; x < maxX; x++) {
						//data[x][y] = (T)(Float)counter;//node.value;
						data[x][y] = new Vector4f(node.value.getPos().x, node.value.getPos().y, node.value.getPos().z, 1.0f);
					}
				}
				
			}
			
			
		}
		
	}
	
	public Vector4f[][] getInputLightRadiusImage() {
		
		Vector4f[][] out = new Vector4f[sx][sy];
		
		for (int y = 0; y < sy; y++) {
			for (int x = 0; x < sx; x++) {
				out[x][y] = new Vector4f(data[x][y].getRadius(), data[x][y].getRadius(), data[x][y].getRadius(), 1.0f);
			}
		}
		
		return out;
	}
	
	/** Generates an image from the quad-tree with a specified limit to the recursion */
	public Vector4f[][] generateRadiusImage(int maximumLevel) {
		
		Vector4f[][] out = new Vector4f[sx][sy];
		
		recursiveGenerationRadius(root, out, maximumLevel, 0, 0, sx, sy);
		
		return out;
	}
	
	/** Recursive method to generate a 2d array representing the quad-tree (the value debug allows the cell borers to be displayed) */
	public void recursiveGenerationRadius(Node node, Vector4f[][] data, int level, int minX, int minY, int maxX, int maxY) {
		
		float scale = 1.0f/1.0f;
		
		if (level > 0 && node != null) {
			
			int centerX = minX + (maxX-minX)/2;
			int centerY = minY + (maxY-minY)/2;
			
			if (node.lt != null && node.lb != null && node.rt != null && node.rb != null) {
				recursiveGenerationRadius(node.lt, data, level-1, minX, minY, centerX, centerY);
				recursiveGenerationRadius(node.lb, data, level-1, minX, centerY, centerX, maxY);
				recursiveGenerationRadius(node.rt, data, level-1, centerX, minY, maxX, centerY);
				recursiveGenerationRadius(node.rb, data, level-1, centerX, centerY, maxX, maxY);
			} else if (debug) {
				for (int x = minX; x < maxX; x++) {
					data[x][maxY-1] = new Vector4f(0.5f, 0.5f, 0.5f, 1.0f);
				}
				for (int y = minY; y < maxY; y++) {
					data[maxX-1][y] = new Vector4f(0.5f, 0.5f, 0.5f, 1.0f);
				}
			}
			
		} else {
			
			//System.out.println(minX + "/" + maxX + ":" + minY + "/" + maxY);
			//counter += 0.05f;
			
			if (debug) {
				
				for (int y = minY; y < maxY-1; y++) {
					for (int x = minX; x < maxX-1; x++) {
						//data[x][y] = (T)(Float)counter;//node.value;
						data[x][y] = new Vector4f(node.value.getRadius()*scale, node.value.getRadius()*scale, node.value.getRadius()*scale, 1.0f);
					}
				}
				
				for (int x = minX; x < maxX; x++) {
					data[x][maxY-1] = new Vector4f(0.5f, 0.5f, 0.5f, 1.0f);//node.value;
				}
				for (int y = minY; y < maxY; y++) {
					data[maxX-1][y] = new Vector4f(0.5f, 0.5f, 0.5f, 1.0f);//node.value;
				}
				
			} else {
				
				for (int y = minY; y < maxY; y++) {
					for (int x = minX; x < maxX; x++) {
						//data[x][y] = (T)(Float)counter;//node.value;
						data[x][y] = new Vector4f(node.value.getRadius()*scale, node.value.getRadius()*scale, node.value.getRadius()*scale, 1.0f);
					}
				}
				
			}
			
			
		}
		
	}
	
	// ---------------------------- DEBUG END ------------------------------------
	
	/**
	 * Comments on bounding sphere:
	 * 
	 * http://en.wikipedia.org/wiki/Smallest-circle_problem
	 * 
	 * https://www10.informatik.uni-erlangen.de/de/Teaching/Courses/SS2008/RBD/bounding_volumes.pdf (approximative algorithm)
	 * 
	 * LP
	 * http://theory.stanford.edu/~megiddo/pdf/lp3.pdf
	 * 
	 * welzl
	 * http://www.gamedev.net/page/resources/_/technical/graphics-programming-and-theory/welzl-r2484
	 * 
	 */
	
	/**
	public void recursiveSum(double[] value, Node node, int level, int minX, int minY, int maxX, int maxY,
			Vector4f[][] directIllumSingle, ImagePlanePosition<Vector4f, Vector3f, Float, Float> lightPlanePositions,
			Vector4f pixelColor, Vector3f pixelPos, Vector3f pixelNormal) {
		
		// check distance
		float distance = Vector3f.sub(pixelPos, node.value.getPos(), null).length();
		
		if (distance > 10 && level > 0 && node != null) {
			
			int centerX = minX + (maxX-minX)/2;
			int centerY = minY + (maxY-minY)/2;
			
			if ((node.lt != null && node.lb != null && node.rt != null && node.rb != null)) {
				recursiveSum(value, node.lt, level-1, minX, minY, centerX, centerY, directIllumSingle, lightPlanePositions, pixelColor, pixelPos, pixelNormal);
				recursiveSum(value, node.lb, level-1, minX, centerY, centerX, maxY, directIllumSingle, lightPlanePositions, pixelColor, pixelPos, pixelNormal);
				recursiveSum(value, node.rt, level-1, centerX, minY, maxX, centerY, directIllumSingle, lightPlanePositions, pixelColor, pixelPos, pixelNormal);
				recursiveSum(value, node.rb, level-1, centerX, centerY, maxX, maxY, directIllumSingle, lightPlanePositions, pixelColor, pixelPos, pixelNormal);
			}
			
		} else {
			
			recursiveSumLeaf(value, minX, minY, maxX, maxY, 
					directIllumSingle, lightPlanePositions, pixelColor, pixelPos, pixelNormal);
			
		}
		
	}
	private void recursiveSumLeaf(double[] newLight, int minX, int minY, int maxX, int maxY, 
			Vector4f[][] directIllumSingle, ImagePlanePosition<Vector4f, Vector3f, Float, Float> lightPlanePositions,
			Vector4f pixelColor, Vector3f pixelPos, Vector3f pixelNormal) {
		
		// over all pixels
		for (int ly = minY; ly < maxY; ly++) {
			for (int lx = minX; lx < maxX; lx++) {
				
				Vector4f lightPixelColor = new Vector4f(directIllumSingle[lx][ly]);
				Vector3f lightPixelPos = lightPlanePositions.getWs_position(lx, ly);
				
				Vector3f lightDir = new Vector3f();
				lightDir.x = lightPixelPos.x - pixelPos.x;
				lightDir.y = lightPixelPos.y - pixelPos.y;
				lightDir.z = lightPixelPos.z - pixelPos.z;
				lightDir.normalise();
				
				float ndotl = Vector3f.dot(lightDir, pixelNormal);
				if (ndotl < 0) {
					ndotl = 0;
				}
				if (ndotl > 1) {
					ndotl = 1;
				}
				
				Vector3f lPPos = new Vector3f(lightPixelPos.x, lightPixelPos.y, lightPixelPos.z);
				Vector3f dst = new Vector3f();
				Vector3f.sub(pixelPos, lPPos, dst);
				double len = (double)dst.length()*1.0;
				double d = 0.0;
				if (len < 100.0) {
					d = 1.0/(len*len);
				}
				if (len < 1) {
					d = 1;
				}
				
				// surface color times light
				newLight[0] += (indirect_factor * d * ndotl * lightPixelColor.x * pixelColor.x);
				newLight[1] += (indirect_factor * d * ndotl * lightPixelColor.y * pixelColor.y);
				newLight[2] += (indirect_factor * d * ndotl * lightPixelColor.z * pixelColor.z);
				
			}
		}
		
	}*/
	
	
	/** Generates the summation over all pixels recursively */
	public double[] generateSummationOverImage(int maximumLevel, float radius_multiplier,
			Vector4f[][] directIllumSingle, 
			ImagePlane<Vector4f, Vector3f, Float, Float> lightPlanes,
			ImagePlanePosition<Vector4f, Vector3f, Float, Float> lightPlanePositions,
			Vector4f pixelColor, Vector4f pixelPos, Vector3f pixelNormal) {
		
		double[] out = new double[]{0.0, 0.0, 0.0};
		
		if (maximumLevel > localMaxLevel) {
			maximumLevel = localMaxLevel;
		}
		
		recursiveSum(radius_multiplier, out, root, maximumLevel, 0, 0, sx, sy, directIllumSingle, lightPlanes, lightPlanePositions, pixelColor, pixelPos, pixelNormal);
		
		out[0] = (float)(out[0]/normalizationFactor);
		out[1] = (float)(out[1]/normalizationFactor);
		out[2] = (float)(out[2]/normalizationFactor);
		
		return out;
	}
	
	/**
	 * Recursion to sum over all pixels.
	 * Requires to store also the positions inside the interval to sum if necessary.
	 * 
	 * The output is double[] value. All summation works on this variable.
	 */
	public void recursiveSum( float radius_multiplier,
			double[] value, Node node, int level, int minX, int minY, int maxX, int maxY,
			Vector4f[][] directIllumSingle,
			ImagePlane<Vector4f, Vector3f, Float, Float> lightPlanes,
			ImagePlanePosition<Vector4f, Vector3f, Float, Float> lightPlanePositions,
			Vector4f pixelColor, Vector4f pixelPos, Vector3f pixelNormal) {
		
		Vector3f bsPos = node.value.getPos(); // bounding-sphere position
		Vector3f pPos = new Vector3f(pixelPos.x, pixelPos.y, pixelPos.z);
		float distance_To_Max_Boundingsphere = Vector3f.sub(bsPos, pPos, null).length();
		
		if (distance_To_Max_Boundingsphere < radius_multiplier * node.value.getRadius()) { // only do if not too far away
		
			if (level > 0 && node != null) {
				
				int centerX = minX + (maxX-minX)/2;
				int centerY = minY + (maxY-minY)/2;
				
				if (node.lt != null && node.lb != null && node.rt != null && node.rb != null) {
					recursiveSum(radius_multiplier, value, node.lt, level-1, minX, minY, centerX, centerY, directIllumSingle, lightPlanes, lightPlanePositions, pixelColor, pixelPos, pixelNormal);
					recursiveSum(radius_multiplier, value, node.lb, level-1, minX, centerY, centerX, maxY, directIllumSingle, lightPlanes, lightPlanePositions, pixelColor, pixelPos, pixelNormal);
					recursiveSum(radius_multiplier, value, node.rt, level-1, centerX, minY, maxX, centerY, directIllumSingle, lightPlanes, lightPlanePositions, pixelColor, pixelPos, pixelNormal);
					recursiveSum(radius_multiplier, value, node.rb, level-1, centerX, centerY, maxX, maxY, directIllumSingle, lightPlanes, lightPlanePositions, pixelColor, pixelPos, pixelNormal);
				}
				
			} else {
				
				//System.out.println("leaf " + node);
				
				recursiveSumLeaf(value, minX, minY, maxX, maxY, 
						directIllumSingle, lightPlanes, lightPlanePositions, pixelColor, pixelPos, pixelNormal);
				
			}
			
		}
		
	}
	
	/**
	 * 
	 * @param newLight ... the value that is incremented (summed over)
	 * @param minX
	 * @param minY
	 * @param maxX
	 * @param maxY
	 * @param directIllumSingle (at i)
	 * @param lightPlanePositions (at i)
	 */
	private void recursiveSumLeaf(double[] newLight, int minX, int minY, int maxX, int maxY, 
			Vector4f[][] directIllumSingle, 
			ImagePlane<Vector4f, Vector3f, Float, Float> lightPlanes,
			ImagePlanePosition<Vector4f, Vector3f, Float, Float> lightPlanePositions,
			Vector4f pixelColor, Vector4f pixelPos, Vector3f pixelNormal) {
		
		// over all pixels
		for (int ly = minY; ly < maxY; ly++) {
			for (int lx = minX; lx < maxX; lx++) {
				
				//System.out.println("inner loop: " + lx + "/" + ly);
				
				Vector4f lightPixelColor = new Vector4f(directIllumSingle[lx][ly]);
				Vector3f lightPixelPos = lightPlanePositions.getWs_position(lx, ly);
				Vector3f lightPixelNormal = new Vector3f(lightPlanes.Normal(lx, ly));
				
				Vector3f lightDir = new Vector3f();
				lightDir.x = lightPixelPos.x - pixelPos.x;
				lightDir.y = lightPixelPos.y - pixelPos.y;
				lightDir.z = lightPixelPos.z - pixelPos.z;
				lightDir.normalise();
				
				/*float ndotl = Vector3f.dot(lightDir, pixelNormal);
				if (ndotl < 0) {
					ndotl = 0;
				}
				if (ndotl > 1) {
					ndotl = 1;
				}
				//ndotl = 1;*/
				
				// only for the image pixel
				float ndotl = Vector3f.dot(lightDir, pixelNormal); //lightPixelNormal
				if (ndotl < 0) {
					ndotl = 0;
				}
				if (ndotl > 1) {
					ndotl = 1;
				}
				
				// for the light pixel
				float ndotl_light = Vector3f.dot((Vector3f)lightDir.negate(), lightPixelNormal);
				if (ndotl_light < 0) {
					ndotl_light = 0;
				}/* else {
					ndotl_light = 1;
				}*/
				if (ndotl_light > 1) {
					ndotl_light = 1;
				}
				//ndotl_light = (float)Math.sqrt(ndotl_light);
				
				Vector3f pPos = new Vector3f(pixelPos.x, pixelPos.y, pixelPos.z);
				Vector3f lPPos = new Vector3f(lightPixelPos.x, lightPixelPos.y, lightPixelPos.z);
				Vector3f dst = new Vector3f();
				Vector3f.sub(pPos, lPPos, dst);
				double len = (double)dst.length();
				double attenuation = 1.0/(1.0 + Math.pow(len, 2.0));
				
				// surface color times light
				newLight[0] += (indirect_factor * attenuation * ndotl * ndotl_light * lightPixelColor.x * pixelColor.x);
				newLight[1] += (indirect_factor * attenuation * ndotl * ndotl_light * lightPixelColor.y * pixelColor.y);
				newLight[2] += (indirect_factor * attenuation * ndotl * ndotl_light * lightPixelColor.z * pixelColor.z);
				
			}
		}
		
	}
	
	
}
